import sys
import numpy as np

from . import utils as outils
from . plot import Plot
from . passband import PassbandContainer, init_bolometric_passband
from .. base.types import INT
from .. binary_system.curves.community import RadialVelocitySystem
from .. import settings
from .. utils import is_empty, jd_to_phase
from .. logger import getLogger
from .. import (
    units as u,
    umpy as up
)
from ..photometric_standards.standards_handlers import load_standard

logger = getLogger('observer.observer')


class Observables(object):
    def __init__(self, observer):
        self.observer = observer

    def lc(self, from_phase=None, to_phase=None, phase_step=None, phases=None, normalize=False,
           from_time=None, to_time=None, time_step=None, times=None, flux_unit=None):
        return self.observer.lc(from_phase, to_phase, phase_step, phases, normalize,
                                from_time, to_time, time_step, times, flux_unit)

    def rv(self, from_phase=None, to_phase=None, phase_step=None, phases=None, normalize=False, method=None,
           from_time=None, to_time=None, time_step=None, times=None):
        return self.observer.rv(from_phase, to_phase, phase_step, phases, normalize, method,
                                from_time, to_time, time_step, times)


class Observer(object):
    def __init__(self, passband=None, system=None):
        """
        The observer class is responsible for the calculation of synthetic observations. Initialization of the Observer
        class instance requires the initialized System instance (SingleSystem, BinarySystem) and in case of light curves
        a (set of) passband(s) in which the observation should be produced.

        List of available passbands is accessible via:
        ::

            from elisa import settings

            settings.PASSBANDS

        This class currently supports calculation of light curves (lc) and radial velocity curves (rv) using functions
        `observer_instance.lc` or `observer_instance.rv` (see documentation for the respective function for
        further details).

        After initialization, the following attributes are available for each instances of Observer class:

            - left_bandwidth, right_bandwidth: the smallest interval of wavelengths encompassing all desired passbands
            - passband: {passband1: PassbandContainer, ...}, dictionary containing the PassbandContainer
              (response curve of the filter) corresponding to each passband.

        :param passband: Union[string, list]; for valid filter name see settings.py file
        :param system: Union[SingleSystem, BinarySystem]; system instance (BinarySystem or SingleSystem)
        """
        if passband is None:
            passband = list()
        logger.info("initialising Observer instance")
        # specifying what kind of system is observed
        self._system = system
        self._system_cls = type(self._system)

        self.left_bandwidth = sys.float_info.max
        self.right_bandwidth = 0.0
        self.passband = dict()
        self.init_passband(passband)

        self._observables = ['lc', 'rv']
        self.phases = None
        self.times = None
        self.fluxes = None
        self.magnitudes = None
        self._flux_unit = u.W / u.m**2
        self.radial_velocities = dict()
        self.rv_unit = None
        self.zero_points = dict(system=None)

        self.plot = Plot(self)
        self.observe = Observables(self)

    @property
    def system_cls(self):
        return self._system_cls

    @system_cls.setter
    def system_cls(self, value):
        self._system_cls = value

    @property
    def flux_unit(self):
        return self._flux_unit

    @flux_unit.setter
    def flux_unit(self, value):
        self._flux_unit = u.Unit(value)

    def init_passband(self, passband):
        """
        Passband initializing method for Observer instance.
        During initialialization `self.passband` Dict is filled in following manner::

            {`passband`: PassbandContainer()}

        and global left and right passband bandwidth are set.
        If case of several passband defined on different intervals, e.g. ([350, 650], [450, 750]), the global limits
        are total boarder values, in this example: [350, 750].

        :param passband: Union[str; Iterable[str]]
        """
        passband = [passband] if isinstance(passband, str) else passband
        for band in passband:
            if band in ['bolometric']:
                psbnd, right_bandwidth, left_bandwidth = init_bolometric_passband()
            else:
                psbnd = PassbandContainer.from_name(passband=band)
                left_bandwidth, right_bandwidth = psbnd.get_bandwidth()

            self.setup_bandwidth(left_bandwidth=left_bandwidth, right_bandwidth=right_bandwidth)
            self.passband[band] = psbnd

    def setup_bandwidth(self, left_bandwidth, right_bandwidth):
        """
        Find whether supplied left and right bandwidth are within currently set boundaries. If not, the supplied
        `left_bandwidth` and `right_bandwidth` are assigned as a new global wavelength boundaries.

        :param left_bandwidth: float;
        :param right_bandwidth: float;
        """
        if left_bandwidth < self.left_bandwidth:
            self.left_bandwidth = left_bandwidth
        if right_bandwidth > self.right_bandwidth:
            self.right_bandwidth = right_bandwidth

    def lc(self, from_phase=None, to_phase=None, phase_step=None, phases=None, normalize=False,
           from_time=None, to_time=None, time_step=None, times=None, flux_unit=None):
        """
        Method for simulated light curve observation. Computes a light curve based on input parameters and the System
        supplied during initialization of the Observer instance. Returns light curves for each passband defined in the
        Observer instance. Times of observations can be supplied in either time or phase domain.

        :param from_time: float; starting time of the observation
        :param to_time: float; end time of the observation
        :param time_step: float; time increment of the observations
        :param times: Iterable float; array of times at which to perform an observations, if this parameter is provided
                                      the suplied 'from_time`, `to_time` `time_step` becomes irrelevant
        :param normalize: bool; if True, the output is normalized to maximum=1
        :param from_phase: float; starting phase of the observation
        :param to_phase: float; end phase of the observation
        :param phase_step: float; phase increment of the observations
        :param phases: Iterable float; array of phases at which to perform an observations, if this parameter is
                                       provided the supplied 'from_phase`, `to_phase` `phase_step` becomes irrelevant
        :param flux_unit: astropy.Units; unit of flux
        :return: Dict;
        """
        if normalize:
            if flux_unit in [None, u.dimensionless_unscaled]:
                self.flux_unit = u.dimensionless_unscaled
            else:
                raise ValueError('You can either produce normalized light curve or specify `flux_unit` other '
                                 'than dimensionless unscaled. Change input parameters.')
        else:
            self.flux_unit = u.Unit(flux_unit) if flux_unit is not None else self.flux_unit

        phases = self.manage_time_series(from_phase, to_phase, phase_step, phases, from_time, to_time, time_step, times)

        # reduce phases to only unique ones from interval (0, 1) in general case without pulsations
        base_phases, base_phases_to_origin = self.phase_interval_reduce(phases)

        logger.info(f"observation is running")
        # calculates lines of sight for corresponding phases
        position_method = self._system.get_positions_method()

        lc_kwargs = dict(
            passband=self.passband,
            left_bandwidth=self.left_bandwidth,
            right_bandwidth=self.right_bandwidth,
            phases=base_phases,
            position_method=position_method
        )

        curves = self._system.compute_lightcurve(**lc_kwargs)

        # remap unique phases back to original phase interval
        for items in curves:
            curves[items] = np.array(curves[items])[base_phases_to_origin]

            # adding additional light
            correction = np.mean(curves[items]) * self._system.additional_light / (1.0 - self._system.additional_light)
            curves[items] += correction

        self.phases = phases + self._system.phase_shift
        if normalize or self.flux_unit == u.dimensionless_unscaled:
            self.fluxes, _ = outils.normalize_light_curve(y_data=curves, kind='maximum', top_fraction_to_average=0.0)
        else:
            curves = outils.adjust_flux_for_distance(curves, self._system.distance)
            if self.flux_unit in [None, u.W / u.m ** 2]:
                self.fluxes = curves
            elif self.flux_unit == u.mag:
                if self.zero_points['system'] != settings.MAGNITUDE_SYSTEM.lower():
                    self.zero_points = load_standard(settings.MAGNITUDE_SYSTEM)
                self.fluxes = outils.convert_to_magnitudes(curves, self.zero_points)
                self.magnitudes = self.fluxes
            else:
                raise ValueError(f'Unknown value for `Observer.flux_unit`: {self.flux_unit}')

        logger.info("observation finished")
        return self.phases, self.fluxes

    def rv(self, from_phase=None, to_phase=None, phase_step=None, phases=None, normalize=False, method=None,
           from_time=None, to_time=None, time_step=None, times=None):
        """
        Method for simulated radial velocity curve observation. Computes a radial velocity curve based on input
        parameters and the System supplied during initialization of the Observer instance. Times of observations
        can be supplied in either time or phase domain.

        :param from_time: float; starting time of the observation
        :param to_time: float; end time of the observation
        :param time_step: float; time increment of the observations
        :param times: Iterable float; array of times at which to perform an observations, if this parameter is provided
                                      the suplied 'from_time`, `to_time` `time_step` becomes irrelevant

        :param from_phase: float; starting phase of the observation
        :param to_phase: float; end phase of the observation
        :param phase_step: float; phase increment of the observations
        :param phases: Iterable float; array of phases at which to perform an observations, if this parameter is
                                       provided the supplied 'from_phase`, `to_phase` `phase_step` becomes irrelevant

        :param normalize: bool;
        :param method: str; method for calculation of radial velocities, `kinematic` or `radiometric`
        :return: Tuple[numpy.array, numpy.array, numpy.array]; phases, primary rv, secondary rv
        """
        method = settings.RV_METHOD if method is None else method

        phases = self.manage_time_series(from_phase, to_phase, phase_step, phases, from_time, to_time, time_step, times)

        # reduce phases to only unique ones from interval (0, 1) in general case without pulsations
        base_phases, base_phases_to_origin = self.phase_interval_reduce(phases)

        self.radial_velocities = self._system.compute_rv(
            **dict(
                phases=base_phases,
                position_method=self._system.get_positions_method(),
                method=method
            )
        )

        # remap unique phases back to original phase interval
        for items in self.radial_velocities:
            self.radial_velocities[items] = np.array(self.radial_velocities[items])[base_phases_to_origin]

        self.phases = phases + self._system.phase_shift
        self.rv_unit = u.m / u.s
        if normalize:
            self.rv_unit = u.dimensionless_unscaled
            _max = np.max([np.max(item) for item in self.radial_velocities.values()])
            self.radial_velocities = {key: value / _max for key, value in self.radial_velocities.items()}

        return self.phases, self.radial_velocities

    def phase_interval_reduce(self, phases):
        """
        Function reduces original phase interval to base interval (0, 1) in case of LC without pulsations.

        :param phases: ndarray; phases to reduce
        :return: Tuple; (base_phase: ndarray, reverse_indices: ndarray)

        ::

            base_phases:  ndarray of unique phases between (0, 1)
            reverse_indices: ndarray mask applicable to `base_phases` which will reconstruct original `phases`
        """
        from ..binary_system.system import BinarySystem
        from ..single_system.system import SingleSystem

        if self._system_cls == BinarySystem or str(self._system_cls) == str(BinarySystem):
            # function shouldn't search for base phases if system has pulsations or is asynchronous with spots
            has_pulsation_test = self._system.primary.has_pulsations() | self._system.secondary.has_pulsations()

            test1 = (self._system.primary.synchronicity != 1.0) & self._system.primary.has_spots()
            test2 = (self._system.secondary.synchronicity != 1.0) & self._system.secondary.has_spots()
            asynchronous_spotty_test = test1 | test2

            if has_pulsation_test | asynchronous_spotty_test:
                return phases, up.arange(phases.shape[0])
            else:
                base_interval = np.round(phases % 1, 9)
                return np.unique(base_interval, return_inverse=True)

        elif self._system_cls == SingleSystem or str(self._system_cls) == str(SingleSystem):
            has_pulsation_test = self._system.star.has_pulsations()
            has_spot_test = self._system.star.has_spots()

            # the most complex case, has to be solved for each phase
            if has_pulsation_test:
                return phases, up.arange(phases.shape[0])
            # in case of just spots on surface, unique (0.1) phases are only needed
            elif has_spot_test and not has_pulsation_test:
                base_interval = np.round(phases % 1, 9)
                return np.unique(base_interval, return_inverse=True)
            # in case of clear surface no pulsations and spots, only single observation is needed
            else:
                return np.zeros(1), np.zeros(phases.shape[0], dtype=INT)

        elif self._system_cls == RadialVelocitySystem or str(self._system_cls) == str(RadialVelocitySystem):
            return phases, up.arange(phases.shape[0], dtype=INT)

        else:
            raise NotImplemented("Not implemented.")

    def manage_time_series(self, from_phase=None, to_phase=None, phase_step=None, phases=None,
                           from_time=None, to_time=None, time_step=None, times=None):
        """
        Converts input time series parameters into photometric phases.

        :param from_phase: float; starting phase of the observation
        :param to_phase: float; end phase of the observation
        :param phase_step: float; phase increment of the observations
        :param phases: Iterable float; array of phases at which to perform an observations, if this parameter is
                                       provided the supplied 'from_phase`, `to_phase` `phase_step` becomes irrelevant
        :param from_time: float; starting time of the observation
        :param to_time: float; end time of the observation
        :param time_step: float; time increment of the observations
        :param times: Iterable float; array of times at which to perform an observations, if this parameter is provided
                                      the suplied 'from_time`, `to_time` `time_step` becomes irrelevant

        """
        phases_supplied = not (phases is None and (from_phase is None or to_phase is None or phase_step is None))
        times_supplied = not (times is None and (from_time is None or to_time is None or time_step is None))

        message = "Please pick arguments either from phase-domain: `from_phase`, 'to_phase`, " \
                  "`phase_step` or `phases` or from time-domain parameters: `from_time`, `to_time`, " \
                  "`time_step` or `times`."
        if not (phases_supplied or times_supplied):
            raise ValueError("Missing arguments.\n" + message)
        if phases_supplied and times_supplied:
            raise ValueError("You specified time series in phase and time domain at once.\n" + message)

        if times_supplied:
            if is_empty(times):
                times = up.arange(start=from_time, stop=to_time, step=time_step)

            phases = jd_to_phase(times, self._system.period, self._system.t0, centre=0.5)
        else:
            if is_empty(phases):
                phases = up.arange(start=from_phase, stop=to_phase, step=phase_step)

        phases = np.array(phases) - self._system.phase_shift

        return phases



import numpy as np

from ... base.curves import rv_point
from . import c_router
from .. orbit.orbit import distance_to_center_of_mass
from ... import settings
from ... import (
    umpy as up,
    units as u
)


def _radial_velocity(semi_major_axis, inclination, eccentricity, argument_of_periastron, period, true_anomaly):
    """
    Compute radial velocity for given paramters.

    :param semi_major_axis: float;
    :param inclination: float;
    :param eccentricity: float;
    :param argument_of_periastron: float;
    :param true_anomaly: float or numpy.array;
    :param period: float;
    :return: Union[float, numpy.array];
    """
    a = 2.0 * up.pi * semi_major_axis * up.sin(inclination)
    b = period * up.sqrt(1.0 - up.power(eccentricity, 2))
    c = up.cos(true_anomaly + argument_of_periastron) + (eccentricity * up.cos(argument_of_periastron))
    return - a * c / b


def kinematic_radial_velocity(binary, **kwargs):
    """
    Calculates radial velocity curves of the `binary` system using radial velocities of centres of masses.

    :param binary: elisa.binary_system.system.BinarySystem; binary system instance
    :param kwargs: Dict;
    :**kwargs options**:
        * **position_method** * -- function that is used to calculate orbital motion
        * **phases** * -- phases in which to calculate

    :return: Dict[str, Unionp[float, numpy.array]; index of values are related to index of phases
    """
    position_method = kwargs.pop("position_method")
    phases = kwargs.pop("phases")
    orbital_motion = position_method(input_argument=phases, return_nparray=True, calculate_from='phase')

    sma_primary, sma_secondary = distance_to_center_of_mass(binary.primary.mass, binary.secondary.mass, 1.0)

    # in base SI units
    sma_primary *= binary.semi_major_axis
    sma_secondary *= binary.semi_major_axis
    period = np.float64((binary.period * u.DefaultBinarySystemUnits.system.period).to(u.TIME_UNIT))

    rv_primary = _radial_velocity(sma_primary, binary.inclination, binary.eccentricity,
                                  binary.argument_of_periastron, period, orbital_motion[:, 3]) * -1.0

    rv_secondary = _radial_velocity(sma_secondary, binary.inclination, binary.eccentricity,
                                    binary.argument_of_periastron, period, orbital_motion[:, 3])

    rvs = {'primary': rv_primary + binary.gamma, 'secondary': rv_secondary + binary.gamma}
    return rvs


def compute_circular_synchronous_rv_curve(binary, **kwargs):
    """
    Compute radial velocity curve for synchronous circular binary system.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param kwargs: Dict;
            * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
            * ** left_bandwidth ** * - float
            * ** right_bandwidth ** * - float
            * ** position_method** * - function definition; to evaluate orbital positions
            * ** phases ** * - numpy.array

    :return: Dict[str, numpy.array];
    """
    initial_system = c_router.prep_initial_system(binary)
    rv_labels = list(settings.BINARY_COUNTERPARTS.keys())
    args = (binary, initial_system, kwargs.pop("phases"), rv_point.compute_rv_at_pos, rv_labels)
    return c_router.produce_circular_sync_curves(*args, **kwargs)


def compute_circular_spotty_asynchronous_rv_curve(binary, **kwargs):
    """
    Function returns rv curve of asynchronous systems with circular orbits and spots.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** - float
        * ** right_bandwidth ** - float
        * ** atlas ** - str

    :return: Dict; rv for each component
    """
    rv_labels = list(settings.BINARY_COUNTERPARTS.keys())
    return c_router.produce_circular_spotty_async_curves(binary, rv_point.compute_rv_at_pos, rv_labels, **kwargs)


def compute_circular_pulsating_rv_curve(binary, **kwargs):
    """
    Function returns rv curve of pulsating systems with circular orbits.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** - float
        * ** right_bandwidth ** - float
        * ** atlas ** - str
        * ** phases ** * - numpy.array

    :return: Dict; rv for each component
    """
    initial_system = c_router.prep_initial_system(binary, **dict(build_pulsations=False))
    rv_labels = list(settings.BINARY_COUNTERPARTS.keys())
    args = (binary, initial_system, kwargs.pop("phases"), rv_point.compute_rv_at_pos, rv_labels)
    return c_router.produce_circular_pulsating_curves(*args, **kwargs)


def compute_eccentric_rv_curve_no_spots(binary, **kwargs):
    """
    General function for generating rv curves of binaries with eccentric orbit and no spots.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** - float
        * ** right_bandwidth ** - float
        * ** atlas ** - str

    :return: Dict; rv for each component
    """
    rv_labels = list(settings.BINARY_COUNTERPARTS.keys())
    return c_router.produce_ecc_curves_no_spots(binary, rv_point.compute_rv_at_pos, rv_labels, **kwargs)


def compute_eccentric_spotty_rv_curve(binary, **kwargs):
    """
    General function for generating rv curves of binaries with eccentric orbit and spots.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** - float
        * ** right_bandwidth ** - float
        * ** atlas ** - str

    :return: Dict; rv for each component
    """
    rv_labels = list(settings.BINARY_COUNTERPARTS.keys())
    return c_router.produce_ecc_curves_with_spots(binary, rv_point.compute_rv_at_pos, rv_labels, **kwargs)



from numba import jit
import numpy as np


@jit(nopython=True, cache=True)
def create_distance_vector_matrix(points1, points2):
    """
    Calculates distances between every point couple in arrays points1 and points2
    :param points1: numpy.array; (a, 3)
    :param points2: numpy.array; (b, 3)
    :return: numpy.array; (a, b, 3)
    """
    result = np.empty((points1.shape[0], points2.shape[0], points1.shape[1]))
    for ii in range(points1.shape[0]):
        for jj in range(points2.shape[0]):
            for kk in range(points2.shape[1]):
                result[ii, jj, kk] = points2[jj, kk] - points1[ii, kk]
    return result


@jit(nopython=True, cache=True)
def calculate_lengths_in_3d_array(matrix):
    """
    Calculates lengths of each 3d vector stored in (a, b, 3) array.

    :param matrix: numpy.array; (a, b, 3)
    :return: numpy.array; (a, b)
    """
    result = np.empty(matrix.shape[:-1])
    for ii in range(matrix.shape[0]):
        for jj in range(matrix.shape[1]):
            result[ii, jj] = (matrix[ii, jj, 0] ** 2 + matrix[ii, jj, 1] ** 2 + matrix[ii, jj, 2] ** 2) ** 0.5

    return result


@jit(nopython=True, cache=True)
def divide_points_in_array_by_constants(matrix, coefficients):
    """
    Divide 3D points in matrix by a separate coefficient (e.g. their norm).

    :param matrix: numpy.array; (a, b, 3)
    :param coefficients: numpy.array; (a, b)
    :return: numpy.array; (a, b, 3)
    """
    result = np.empty(matrix.shape)
    for ii in range(matrix.shape[0]):
        for jj in range(matrix.shape[1]):
            for kk in range(matrix.shape[2]):
                result[ii, jj, kk] = matrix[ii, jj, kk] / coefficients[ii, jj]

    return result



import re

import numpy as np
from scipy.interpolate import interp1d
from copy import deepcopy
from emcee.autocorr import integrated_time, function_1d

from . import shared
from . mixins import MCMCMixin
from .. models.lc import synthetic_binary
from .. models.rv import central_rv_synthetic
from .. params import parameters, conf
from ... import units as u
from ... observer.utils import normalize_light_curve
from ... binary_system import t_layer
from ... binary_system.system import BinarySystem
from ... graphic.mcmc_graphics import Plot as MCMCPlot
from ... observer.observer import Observer
from ... graphic import graphics
from ... logger import getLogger
from elisa.utils import is_empty


logger = getLogger('analytics.binary_fit.plot')

PLOT_UNITS = {
    'system@asini': u.solRad,
    'system@argument_of_periastron': u.degree,
    'system@gamma': u.km/u.s,
    'system@primary_minimum_time': u.d
}


class MCMCPlotMixin(object):
    """
    Graphics module for visualization of MCMC sampling results.
    """
    fit = None

    def corner(self, flat_chain=None, variable_labels=None, normalization=None,
               quantiles=None, truths=False, show_titles=True, plot_units=None, sigma_clip=False, sigma=5, n_bins=20,
               **kwargs):
        """
        Plots complete correlation plot from supplied parameters. Useful only for visualizing the posterior distribution
        of the MCMC samples.

        :param sigma_clip: bool; if True, posterior distribution is cropped within
                                 (mean - `sigma` * std, mean + `sigma` * std) to filter out the outliers,
        :param sigma: float; positive, with of the sigma cropping interval in the multiples of standard deviation of
                             the chain distribution
        :param flat_chain: numpy.array; flattened chain of all parameters, use only if you want display your own chain.
                                        By default, internal `flat_chain` attribute will be used as a source of the
                                        posterior distribution.
        :param variable_labels: List; list of variables during a MCMC run, which is used to identify columns in
                                      `flat_chain`, use only if you want display your own chain
        :param normalization: Dict[str, Tuple(float, float)]; {var_name: (min_boundary, max_boundary), ...} dictionary
                                                              of boundaries defined by user for each variable needed
                                                              to reconstruct real values from normalized `flat_chain`,
                                                              use only if you want display your own chain
        :param quantiles: Iterable; A list of fractional quantiles to show on the 1-D histograms as vertical dashed
                                   lines.
        :param truths: Union[bool, List]; if True, fit results are used to indicate position of found values. If False,
                                          none are shown. If list is supplied, it functions the same
                                          as in corner.corner function
        :param show_titles: bool; If True, labels above histogram with name of the variable, value, error and units
                                  are displayed
        :param plot_units: Dict; Units in which to display the output {variable_name: unit, ...}
        :param n_bins: int; positive, number of bins in each histogram
        """
        corner(self.fit, flat_chain=flat_chain, variable_labels=variable_labels, normalization=normalization,
               quantiles=quantiles, truths=truths, show_titles=show_titles, plot_units=plot_units,
               sigma_clip=sigma_clip, sigma=sigma, n_bins=n_bins, **kwargs)

    def autocorrelation(self, correlations_to_plot=None, flat_chain=None, variable_labels=None):
        """
        Plots correlation function for the output MCMC chain.

        :param correlations_to_plot: List; names of variables (in flat format e.g. system@inclination) which
                                           autocorrelation function will be displayed
        :param flat_chain: numpy.array; optional, flattened chain of all parameters. If None, internal `flat_chain`
                                        attribute will be used as a source of the posterior distribution.
        :param variable_labels: List; list of variables during a MCMC run, which is used to identify columns in
                                      `flat_chain`. Use only if `flat_chain` is not None.
        """
        autocorrelation(self.fit, correlations_to_plot, flat_chain, variable_labels)

    def traces(self, traces_to_plot=None, flat_chain=None, variable_labels=None,
               normalization=None, plot_units=None, truths=False):
        """
        Plots traces of the MCMC samples.

        :param traces_to_plot: List; names of variables which traces will be displayed
        :param flat_chain: numpy.array; optional, flattened chain of all parameters. If None, internal `flat_chain`
                                        attribute will be used as a source of the posterior distribution.
        :param variable_labels: List; list of variables during a MCMC run, which is used to identify columns in
                                      `flat_chain`. Use only if `flat_chain` is not None.
        :param normalization: Dict[str, Tuple(float, float)]; {var_name: (min_boundary, max_boundary), ...} dictionary
                                                              of boundaries defined by user for each variable
                                                              needed to reconstruct real values from normalized
                                                              `flat_chain`, use only if you want display your own chain
        :param plot_units: Dict; Units in which to display the output {variable@name: unit, ...}
        :param truths: bool; if True, fit results are used to indicate position of found values. If False,
                             none are shown. It will not work with a custom chain. (if `flat_chain` is not None).
        """
        traces(self.fit, traces_to_plot=traces_to_plot, flat_chain=flat_chain, variable_labels=variable_labels,
               normalization=normalization, plot_units=plot_units, truths=truths)


class RVPlot(object):
    def __init__(self, instance, data):
        self.fit = instance
        self.data = data

    def model(self, start_phase=-0.6, stop_phase=0.6, number_of_points=300, y_axis_unit=u.km / u.s,
              return_figure_instance=False, **kwargs):
        """
        Plots the RV model described by fit params or calculated by last run of fitting procedure.

        :param start_phase: float; initial orbital phase of the synthetic observations
        :param stop_phase: float; final orbital phase of the synthetic observations
        :param number_of_points: int; number of model points in the synthetic data
        :param y_axis_unit: astropy.unit.Unit;
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        :param kwargs: Dict;
        :**kwargs options for mcmc**:
            * **fit_result** * - Dict - {result_parameter: {value: float, unit: astropy.unit.Unit,
                                         ...(fitting method dependent)}
        """
        logger.debug('Producing/retrieving data for RV plot.')
        plot_result_kwargs = dict()
        fit_result = kwargs.get('fit_result', self.fit.result)

        if fit_result is None:
            raise ValueError('You did not performed radial velocity fit on this instance '
                             'or you did not provided result parameter dictionary.')

        # converting to phase space if necessary
        x_data, y_data, y_err = dict(), dict(), dict()

        for component, data in self.data.items():

            if data.x_unit is u.dimensionless_unscaled:
                x_data[component] = t_layer.adjust_phases(phases=data.x_data, centre=0.0)
            else:
                x_data[component] = t_layer.jd_to_phase(fit_result['system']['primary_minimum_time']['value'],
                                                        fit_result['system']['period']['value'],
                                                        data.x_data, centre=0.0)

            y_data[component] = (data.y_data * data.y_unit).to(y_axis_unit).value
            y_err[component] = (data.y_err * data.y_unit).to(y_axis_unit).value if data.y_err is not None else None

        x_data, y_data, y_err = \
            shared.extend_observations_to_desired_interval(start_phase, stop_phase, x_data, y_data, y_err)

        plot_result_kwargs.update({
            'x_data': x_data,
            'y_data': y_data,
            'y_err': y_err,
            'y_unit': y_axis_unit,
        })

        kwargs_to_replot = parameters.deserialize_result(fit_result)
        kwargs_to_replot = {key: val["value"] for key, val in kwargs_to_replot.items()}
        if 'system@primary_minimum_time' in kwargs_to_replot.keys():
            del kwargs_to_replot['system@primary_minimum_time']
        synth_phases = np.linspace(start_phase, stop_phase, number_of_points)
        rv_fit = central_rv_synthetic(synth_phases, Observer(), **kwargs_to_replot)
        rv_fit = {component: (data * u.VELOCITY_UNIT).to(y_axis_unit).value for component, data in rv_fit.items()}

        interp_fn = {component: interp1d(synth_phases, rv_fit[component])
                     for component in self.data.keys()}
        residuals = {component: y_data[component] - interp_fn[component](x_data[component])
                     for component in self.data.keys()}

        plot_result_kwargs.update({
            'return_figure_instance': return_figure_instance,
            'synth_phases': synth_phases,
            'rv_fit': rv_fit,
            'residuals': residuals,
            'y_unit': y_axis_unit
        })

        logger.debug('Sending data to matplotlib interface.')
        return graphics.binary_rv_fit_plot(**plot_result_kwargs)


class RVPlotLsqr(RVPlot):
    def __init__(self, instance, data):
        super().__init__(instance, data)


class RVPlotMCMC(RVPlot, MCMCPlotMixin):
    def __init__(self, instance, data):
        super().__init__(instance, data)


class LCPlot(object):
    """
    Graphics functions for visualization of LC fit result.
    """
    def __init__(self, instance, data):
        self.fit = instance
        self.data = data

    def model(self, start_phase=-0.6, stop_phase=0.6, number_of_points=300, discretization=5,
              separation=0.1, data_frac_to_normalize=0.1, normalization_kind='maximum', plot_legend=True, loc=1,
              return_figure_instance=False, rasterize=None, **kwargs):
        """
        Prepares data for plotting the model described by fit params or calculated by last run of fitting procedure.

        :param separation: float; separation between different filters, useful while plotting normalized LCs in
                                  different passbands
        :param start_phase: float; initial orbital phase of the synthetic observations
        :param stop_phase: float; final orbital phase of the synthetic observations
        :param number_of_points: int; number of model points in the synthetic data
        :param discretization: float; discretization factor for the primary component during calculation of the
                                      synthetic observations
        :param data_frac_to_normalize: float; optional, between (0, 1), fraction of data points with the highest flux
                                              used for normalization, depends on level of noise in your data
        :param normalization_kind: str; `average` or `maximum`
        :param loc: int; location of the legend
        :param plot_legend: bool; display legend
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        :param rasterize: if True, figure is returned in rasterized form, thus reducing the size of the image
        :param kwargs: Dict;
        :**kwargs options for mcmc**:
            * **fit_result** * - Dict - {result_parameter: {value: float, unit: astropy.unit.Unit,
                                        ...(fitting method dependent)}
        """
        logger.debug('Producing/retrieving data for LC plot.')
        average_kind = normalization_kind
        plot_result_kwargs = dict()
        fit_result = kwargs.get('fit_result', self.fit.result)

        if fit_result is None:
            raise ValueError('You did not performed light curve fit on this instance '
                             'or you did not provided parameter dictionary.')

        # converting to phase space if necessary
        x_data, y_data, y_err = dict(), dict(), dict()

        for band, data in self.data.items():
            if data.x_unit is u.dimensionless_unscaled:
                x_data[band] = t_layer.adjust_phases(phases=data.x_data, centre=0.0)
            else:
                x_data[band] = t_layer.jd_to_phase(fit_result['system']['primary_minimum_time']['value'],
                                                   fit_result['system']['period']['value'],
                                                   data.x_data, centre=0.0)
            y_data[band] = data.y_data
            y_err[band] = data.y_err

        # normalize
        y_data, y_err = normalize_light_curve(y_data, y_err, kind=average_kind,
                                              top_fraction_to_average=data_frac_to_normalize)

        # add y separation to normalized data
        i, y_len = 0, len(y_data)
        for curve in y_data.values():
            curve -= separation * (i - int(y_len/2))
            i += 1

        # extending observations to desired phase interval
        for band, curve in self.data.items():
            phases_extended = np.concatenate((x_data[band] - 1.0, x_data[band], x_data[band] + 1.0))
            phases_extended_filter = np.logical_and(start_phase < phases_extended,  phases_extended < stop_phase)
            x_data[band] = phases_extended[phases_extended_filter]

            y_data[band] = np.tile(y_data[band], 3)[phases_extended_filter]
            if not is_empty(y_err[band]):
                y_err[band] = np.tile(y_err[band], 3)[phases_extended_filter]

        x_data, y_data, y_err = \
            shared.extend_observations_to_desired_interval(start_phase, stop_phase, x_data, y_data, y_err)

        plot_result_kwargs.update({
            'x_data': x_data,
            'y_data': y_data,
            'y_err': y_err,
        })

        kwargs_to_replot = parameters.deserialize_result(fit_result)
        kwargs_to_replot = {key: val["value"] for key, val in kwargs_to_replot.items()}
        if 'system@primary_minimum_time' in kwargs_to_replot.keys():
            del kwargs_to_replot['system@primary_minimum_time']

        # TODO: check this
        kwargs_to_replot = parameters.extend_json_with_atm_params(
            kwargs_to_replot,
            atmosphere_model=self.fit.fit_method_instance.atmosphere_model,
            limb_darkening_coefficients=self.fit.fit_method_instance.limb_darkening_coefficients
        )

        synth_phases = np.linspace(start_phase, stop_phase, number_of_points)
        observer = Observer(passband=self.data.keys(), system=None)
        observer._system_cls = BinarySystem

        lc_fit = synthetic_binary(synth_phases, discretization, observer, **kwargs_to_replot)
        lc_fit, _ = normalize_light_curve(lc_fit, kind=average_kind, top_fraction_to_average=0.001)

        idx = 0
        for curve in lc_fit.values():
            curve -= separation * (idx - int(y_len/2))
            idx += 1

        # interpolating synthetic curves to observations and its residuals
        interp_fn = {band: interp1d(synth_phases, lc_fit[band], kind='cubic') for band in self.data.keys()}
        residuals = {band: y_data[band] - np.mean(y_data[band]) - interp_fn[band](x_data[band]) +
                           np.mean(interp_fn[band](x_data[band])) for band in self.data.keys()}

        plot_result_kwargs.update({
            'return_figure_instance': return_figure_instance,
            'synth_phases': synth_phases,
            'lcs': lc_fit,
            'residuals': residuals,
            'legend': plot_legend,
            'loc': loc,
            'rasterize': rasterize
        })

        logger.debug('Sending data to matplotlib interface.')
        return graphics.binary_lc_fit_plot(**plot_result_kwargs)


class LCPlotLsqr(LCPlot):
    def __init__(self, instance, data):
        super().__init__(instance, data)


class LCPlotMCMC(LCPlot, MCMCPlotMixin):
    def __init__(self, instance, data):
        super().__init__(instance, data)


def serialize_plot_labels(variable_labels):
    """
    Return Tex compatible labels of model parameters.

    :param variable_labels: List; flat format labels of model parameters, e.g. system@inclination
    :return: List; plot labels
    """
    labels = []
    for lbl in variable_labels:
        lbl_s = lbl.split(conf.PARAM_PARSER)

        if re.search(r"|".join(conf.COMPOSITE_FLAT_PARAMS), lbl):
            labels.append(f'{lbl_s[-2]} {conf.PARAMS_KEY_TEX_MAP[lbl_s[-1]]}')
        else:
            labels.append(conf.PARAMS_KEY_TEX_MAP[lbl])
    return labels


def corner(mcmc_fit_instance, flat_chain=None, variable_labels=None, normalization=None,
           quantiles=None, truths=False, show_titles=True, plot_units=None, sigma_clip=False, sigma=5, n_bins=20,
           **kwargs):
    """
    Plots complete correlation plot from supplied parameters. Usefull only for MCMC method

    :param sigma: float; positive, with of the sigma cropping interval in the multiples of standard deviation of
                         the chain distribution
    :param sigma_clip: bool; if True, posterior distribution is cropped within
                             (mean - `sigma` * std, mean + `sigma` * std) to filter out the outliers,
    :param mcmc_fit_instance: Union[elisa.analytics.binary_fit.lc_fit.LCFitMCMC,
                                    elisa.analytics.binary_fit.rv_fit.RVFitMCMC];
    :param flat_chain: numpy.array; flattened chain of all parameters, use only if you want display your own chain.
                                    By default, internal `flat_chain` attribute will be used as a source of the
                                    posterior distribution.
    :param variable_labels: List; list of variables during a MCMC run, which is used to identify columns in
                                  `flat_chain`, use only if you want display your own chain
    :param normalization: Dict[str, Tuple(float, float)]; {var_name: (min_boundary, max_boundary), ...} dictionary
                                                          of boundaries defined by user for each variable needed to
                                                          reconstruct real values from normalized `flat_chain`,
                                                          use only if you want display your own chain
    :param quantiles: Iterable; A list of fractional quantiles to show on the 1-D histograms
                                as vertical dashed lines.
    :param truths: Union[bool, list]; if True, fit results are used to indicate position of found values. If False,
                                      none are shown. If list is supplied, it functions
                                      the same as in corner.corner function.
    :param show_titles: bool; If True, labels above histogram with name of the variable, value,
                              errors and units are displayed
    :param plot_units: Dict; Units in which to display the output {variable_name: unit, ...}
    :param n_bins: int; positive, number of bins in each histogram
    """
    logger.debug('Producing/retrieving data for corner plot.')
    flat_chain = deepcopy(mcmc_fit_instance.flat_chain) if flat_chain is None else deepcopy(flat_chain)
    variable_labels = mcmc_fit_instance.variable_labels if variable_labels is None else variable_labels
    normalization = mcmc_fit_instance.normalization if normalization is None else normalization
    quantiles = [0.16, 0.5, 0.84] if quantiles is None else quantiles
    flat_result = deepcopy(mcmc_fit_instance.flat_result)

    corner_plot_kwargs = dict()
    if flat_chain is None:
        raise ValueError('You can use corner plot after running mcmc method or after loading the flat chain.')

    plot_labels = serialize_plot_labels(variable_labels)
    flat_chain = MCMCMixin.renormalize_flat_chain(flat_chain, mcmc_fit_instance.variable_labels, variable_labels,
                                                  normalization)

    # # transforming units and rearanging them into the correct order
    flat_chain_reduced = np.empty((flat_chain.shape[0], len(variable_labels)))
    plot_units = PLOT_UNITS if plot_units is None else plot_units
    for ii, lbl in enumerate(variable_labels):
        if lbl in plot_units.keys():
            unt = u.Unit(flat_result[lbl]['unit'])
            flat_chain_reduced[:, ii] = (flat_chain[:, ii] * unt).to(plot_units[lbl]).value
            flat_result[lbl]['value'] = (flat_result[lbl]['value'] * unt).to(plot_units[lbl]).value
            flat_result[lbl]["confidence_interval"]['min'] = \
                (flat_result[lbl]["confidence_interval"]['min'] * unt).to(plot_units[lbl]).value
            flat_result[lbl]["confidence_interval"]['max'] = \
                (flat_result[lbl]["confidence_interval"]['max'] * unt).to(plot_units[lbl]).value
            flat_result[lbl]['unit'] = plot_units[lbl].to_string()
        else:
            flat_chain_reduced[:, ii] = flat_chain[:, ii]

    truths = [flat_result[lbl]['value'] for lbl in variable_labels] if truths is True else None

    if sigma_clip:
        for ii, lbl in enumerate(variable_labels):
            tol = 0.5 * sigma * np.abs(flat_result[lbl]["confidence_interval"]['max'] -
                                       flat_result[lbl]["confidence_interval"]['min'])
            mask = np.logical_and(flat_chain_reduced[:, ii] > flat_result[lbl]['value'] - tol,
                                  flat_chain_reduced[:, ii] < flat_result[lbl]['value'] + tol)
            flat_chain_reduced = flat_chain_reduced[mask]

    corner_plot_kwargs.update({
        'flat_chain': flat_chain_reduced,
        'truths': truths,
        'variable_labels': variable_labels,
        'labels': plot_labels,
        'quantiles': quantiles,
        'show_titles': show_titles,
        'fit_params': flat_result,
        'bins': n_bins
    })
    corner_plot_kwargs.update(**kwargs)
    logger.debug('Sending data to matplotlib interface.')
    MCMCPlot.corner(**corner_plot_kwargs)


def autocorrelation(mcmc_fit_instance, correlations_to_plot=None, flat_chain=None, variable_labels=None):
    """
    Plots correlation function of defined parameters.

    :param mcmc_fit_instance: Union[elisa.analytics.binary_fit.lc_fit.LCFitMCMC,
                                    elisa.analytics.binary_fit.rv_fit.RVFitMCMC];
    :param correlations_to_plot: List; names of variables (in flat format e.g. system@inclination) which
                                       autocorrelation function will be displayed
    :param flat_chain: numpy.array; optional, flattened chain of all parameters. If None, internal `flat_chain`
                                    attribute will be used as a source of the posterior distribution.
    :param variable_labels: List; list of variables during a MCMC run, which is used to identify columns in
                                  `flat_chain`. Use only if `flat_chain` is not None.
    """
    autocorr_plot_kwargs = dict()

    flat_chain = deepcopy(mcmc_fit_instance.flat_chain) if flat_chain is None else deepcopy(flat_chain)
    variable_labels = mcmc_fit_instance.variable_labels if variable_labels is None else variable_labels
    correlations_to_plot = variable_labels if correlations_to_plot is None else correlations_to_plot

    if flat_chain is None:
        raise ValueError('You can use trace plot only in case of mcmc method '
                         'or for some reason the flat chain was not found.')

    labels = serialize_plot_labels(variable_labels)

    autocorr_fns = np.empty((flat_chain.shape[0], len(variable_labels)))
    autocorr_time = np.empty((flat_chain.shape[0]))

    for i, lbl in enumerate(variable_labels):
        autocorr_fns[:, i] = function_1d(flat_chain[:, i])
        autocorr_time[i] = integrated_time(flat_chain[:, i], quiet=True)

    autocorr_plot_kwargs.update({
        'correlations_to_plot': correlations_to_plot,
        'autocorr_fns': autocorr_fns,
        'autocorr_time': autocorr_time,
        'variable_labels': variable_labels,
        'labels': labels
    })

    MCMCPlot.autocorr(**autocorr_plot_kwargs)


def traces(mcmc_fit_instance, traces_to_plot=None, flat_chain=None, variable_labels=None,
           normalization=None, plot_units=None, truths=False):
    """
    Plots traces of defined parameters.

    :param mcmc_fit_instance: Union[elisa.analytics.binary_fit.lc_fit.LCFitMCMC,
                                    elisa.analytics.binary_fit.rv_fit.RVFitMCMC];
    :param traces_to_plot: List; names of variables which traces will be displayed
    :param flat_chain: numpy.array; optional, flattened chain of all parameters. If None, internal `flat_chain`
                                    attribute will be used as a source of the posterior distribution.
    :param variable_labels: List; list of variables during a MCMC run, which is used to identify columns in
                                  `flat_chain`. Use only if `flat_chain` is not None.
    :param normalization: Dict[str, Tuple(float, float)]; {var_name: (min_boundary, max_boundary), ...} dictionary
                                                          of boundaries defined by user for each variable needed to
                                                          reconstruct real values from normalized `flat_chain`,
                                                          use only if you want display your own chain
    :param plot_units: Dict; Units in which to display the output {variable_name: unit, ...}
    :param truths: bool; if True, fit results are used to indicate position of found values. If False,
                         none are shown. It will not work with a custom chain. (if `flat_chain` is not None).
    """
    logger.debug('Producing/retrieving data for traces plot.')
    traces_plot_kwargs = dict()

    variable_labels = mcmc_fit_instance.variable_labels if variable_labels is None else variable_labels
    normalization = mcmc_fit_instance.normalization if normalization is None else normalization
    flat_chain = deepcopy(mcmc_fit_instance.flat_chain) if flat_chain is None else deepcopy(flat_chain)
    flat_result = deepcopy(mcmc_fit_instance.flat_result)

    if flat_chain is None:
        raise ValueError('You can use trace plot only in case of mcmc method '
                         'or for some reason the flat chain was not found.')

    flat_chain = MCMCMixin.renormalize_flat_chain(flat_chain, mcmc_fit_instance.variable_labels, variable_labels,
                                                  normalization)
    labels = serialize_plot_labels(variable_labels)

    # transforming units
    plot_units = PLOT_UNITS if plot_units is None else plot_units
    for ii, lbl in enumerate(variable_labels):
        if lbl in plot_units.keys():
            unt = u.Unit(flat_result[lbl]['unit'])
            flat_chain[:, ii] = (flat_chain[:, ii] * unt).to(plot_units[lbl]).value
            flat_result[lbl]['value'] = (flat_result[lbl]['value'] * unt).to(plot_units[lbl]).value
            flat_result[lbl]["confidence_interval"]['min'] = \
                (flat_result[lbl]["confidence_interval"]['min'] * unt).to(plot_units[lbl]).value
            flat_result[lbl]["confidence_interval"]['max'] = \
                (flat_result[lbl]["confidence_interval"]['max'] * unt).to(plot_units[lbl]).value
            flat_result[lbl]['unit'] = plot_units[lbl].to_string()

    traces_to_plot = variable_labels if traces_to_plot is None else traces_to_plot

    traces_plot_kwargs.update({
        'traces_to_plot': traces_to_plot,
        'flat_chain': flat_chain,
        'variable_labels': variable_labels,
        'fit_params': flat_result,
        'truths': truths,
        'labels': labels,
    })

    logger.debug('Sending data to matplotlib interface.')
    MCMCPlot.paramtrace(**traces_plot_kwargs)



# DO NOT CHANGE THIS!!!
# data from Yoon et al., 2009, AJ, 708, 1
VEGA = {
    "system": {
        "inclination": "4.975 deg",
        "rotation_period": "0.604 d",
        "gamma": "-13000 m/s",
        "reference_time": 0.0,
        "phase_shift": 0.0,
        "distance": "10 pc"
    },
    "star": {
        "mass": "2.135 solMass",
        "t_eff": "10059 K",
        "metallicity": -0.39,
        "equivalent_radius": "2.657 solRad"
    }
}



from abc import abstractmethod
from typing import Union

from ... logger import getLogger

from .. params.parameters import BinaryInitialParameters
from .. params import parameters
from .. params.result_handler import FitResultHandler
from . summary import fit_lc_summary_with_error_propagation, simple_lc_fit_summary
from . import least_squares
from . import mcmc
from . import io_tools


logger = getLogger('analytics.binary_fit.lc_fit')

DASH_N = 126


class LCFit(FitResultHandler):
    """
    Class with common methods used during an LC fit.
    """
    def __init__(self, morphology, atmosphere_model, limb_darkening_coefficients):
        super().__init__()
        self.morphology = morphology
        self.fit_method_instance: Union[LCFitLeastSquares, LCFitMCMC, None] = None
        self.atmosphere_model: Union[dict, None] = None
        self.limb_darkening_coefficients: Union[dict, None] = None

    def coefficient_of_determination(self, model_parameters, data, discretization, interp_treshold):
        """
        Function returns R^2 for given model parameters and observed data.

        :param model_parameters: Dict; set of model parameters in json format
        :param data: Dict[str; LCData]; observational data in each passband
        :param discretization: float; discretization factor for the primary component
        :param interp_treshold: int; a number of observation points above which the synthetic curves will be calculated
                                     using `interp_treshold` equally spaced points that will be subsequently
                                     interpolated to the desired times of observation
        :return: float; coefficient of determination (1.0 means a perfect fit to the observations)
        """
        b_parameters = parameters.BinaryInitialParameters(**model_parameters)
        b_parameters.validate_lc_parameters(morphology=self.morphology)
        args = model_parameters, data, discretization, interp_treshold
        return self.fit_method_instance.coefficient_of_determination(*args)

    @abstractmethod
    def resolve_fit_cls(self, morphology: str):
        pass


class LCFitMCMC(LCFit):
    """
    Class for LC fitting using the MCMC method.
    """
    def __init__(self, morphology, atmosphere_model, limb_darkening_coefficients):
        super().__init__(morphology, atmosphere_model, limb_darkening_coefficients)
        self.fit_method_instance = self.resolve_fit_cls(morphology)()
        self.fit_method_instance.atmosphere_model = atmosphere_model
        self.fit_method_instance.limb_darkening_coefficients = limb_darkening_coefficients

        self.flat_chain = None
        self.flat_chain_path = None
        self.normalization = None
        self.variable_labels = None

    def filter_chain(self, **boundaries):
        """
        Filtering MCMC chain down to given parameter intervals. This function is useful in case of bimodal distribution
        of the MCMC chain.

        :param boundaries: Dict; dictionary of boundaries e.g. {'primary@te_ff': (5000, 6000), other parameters ...}
        :return: numpy.array; filtered flat chain
        """
        return io_tools.filter_chain(self, **boundaries)

    def fit(self, x0: BinaryInitialParameters, data, **kwargs):
        """
        Perform MCMC sampling on LCFitMCMC instance.

        :param x0: BinaryInitialParameters; initial info about the model parameters such as status
                                            (fixed, variable, constrained), bounds (prior distribution) and
                                            initial value
        :param data: Dict[LCData]; observational data (light curves in multiple filters)
        :param kwargs: Dict; arguments passed to the fitting method (see AnalyticsTask.fit kwargs for MCMC or
                             mcmc.LightCurveFit.fit for further info)
        :return: Dict; optimized model parameters in JSON format
        """
        x0.validate_lc_parameters(morphology=self.morphology)
        self.result = self.fit_method_instance.fit(data=data, x0=x0, **kwargs)
        self.flat_result = self.fit_method_instance.flat_result

        self.flat_chain = self.fit_method_instance.last_sampler.get_chain(flat=True)
        self.flat_chain_path = self.fit_method_instance.flat_chain_path
        self.normalization = self.fit_method_instance.normalization
        self.variable_labels = list(self.fit_method_instance.fitable.keys())

        logger.info('Fitting and processing of results finished successfully.')
        self.fit_summary()

    def fit_summary(self, filename=None, **kwargs):
        """
        Function produces detailed summary of the current LC fitting task with the possibility to propagate
        uncertainties of the fitted binary model parameters if `propagate_errors` is True.

        :param filename: str; path, place to store summary
        :param kwargs: Dict;
        :**kwargs options**:
            * :propagate_errors: bool; errors of fitted parameters will be propagated to the rest of EB
                                       parameters (takes a while to calculate)
            * :percentiles: List; percentiles used to evaluate confidence intervals from posterior
                                  distribution of EB parameters in MCMC chain . Used only when if
                                  `propagate_errors` is True.
            * :dimensionless_radii: bool; if True (default), radii are provided in SMA, otherwise solRad are used

        """
        propagate_errors, percentiles = kwargs.get('propagate_errors', False), kwargs.get('percentiles', [16, 50, 84])
        dimensionless_radii = kwargs.get('dimensionless_radii', True)
        if not propagate_errors:
            simple_lc_fit_summary(self, filename, dimensionless_radii=dimensionless_radii)
            return

        fit_lc_summary_with_error_propagation(self, filename, percentiles, dimensionless_radii=dimensionless_radii)

        return self.result

    def load_chain(self, filename, discard=0, percentiles=None):
        """
        Function loads MCMC chain along with auxiliary data from json file created after each MCMC run.

        :param percentiles: List; percentile intervals used to generate confidence intervals, provided in form:
                                  [percentile for lower bound of confidence interval, percentile of the centre,
                                  percentile for the upper bound of confidence interval]
        :param discard: int; Discard the first discard steps in the chain as a part of the thermalization phase
                             (default: 0).
        :param filename: str; chain identificator or filename (ending with .json) containing the chain
        :return: Tuple[numpy.ndarray, List, Dict]; flattened mcmc chain, labels of variables in `flat_chain` columns,
                                                   {var_name: (min_boundary, max_boundary), ...} dictionary of
                                                   boundaries defined by user for each variable needed
                                                   to reconstruct real values from normalized `flat_chain` array
        """
        return io_tools.load_chain(self, filename, discard, percentiles)

    def resolve_fit_cls(self, morphology: str):
        """
        Function returns MCMC fitting class suitable for the model based on its morphology.

        :param morphology: str; `detached` or `over-contact`
        :return: Union[mcmc.DetachedLightCurveFit, mcmc.OvercontactLightCurveFit]
        """
        _cls = {"detached": mcmc.DetachedLightCurveFit,
                "over-contact": mcmc.OvercontactLightCurveFit,
                "overcontact": mcmc.OvercontactLightCurveFit}
        return _cls[morphology]


class LCFitLeastSquares(LCFit):
    """
    Class for LC fitting using the Least-Squares method.
    """
    def __init__(self, morphology, atmosphere_model, limb_darkening_coefficients):
        super().__init__(morphology, atmosphere_model, limb_darkening_coefficients)
        self.fit_method_instance = self.resolve_fit_cls(morphology)()
        self.fit_method_instance.atmosphere_model = atmosphere_model
        self.fit_method_instance.limb_darkening_coefficients = limb_darkening_coefficients

    def fit(self, x0: BinaryInitialParameters, data, **kwargs):
        """
        Perform Least-Squares optimization on LCFitLeastSquares instance.

        :param x0: BinaryInitialParameters; initial info about the model parameters such as status
                                            (fixed, variable, constrained), bounds (prior distribution) and
                                            initial value
        :param data: Dict[LCData]; observational data (light curves in multiple filters)
        :param kwargs: Dict; arguments passed to the fitting method (see AnalyticsTask.fit kwargs for Least-Squares or
                             least_squares.LightCurveFit.fit for further info)
        :return: Dict; optimized model parameters in JSON format
        """
        x0.validate_lc_parameters(morphology=self.morphology)
        self.result = self.fit_method_instance.fit(data=data, x0=x0, **kwargs)
        self.flat_result = self.fit_method_instance.flat_result
        logger.info('Fitting and processing of results finished successfully.')
        self.fit_summary()
        return self.result

    def fit_summary(self, path=None, **kwargs):
        """
        Function produces detailed summary of the current LC fitting task.

        :param path: str; path, place to store summary
        :param kwargs:
        :return:
        """
        simple_lc_fit_summary(self, path)

    def resolve_fit_cls(self, morphology: str):
        """
        Function returns Least-Squares fitting class suitable for the model based on its morphology.

        :param morphology: str; `detached` or `over-contact`
        :return: Union[least_squares.DetachedLightCurveFit, least_squares.OvercontactLightCurveFit]
        """
        _cls = {"detached": least_squares.DetachedLightCurveFit,
                "over-contact": least_squares.OvercontactLightCurveFit,
                "overcontact": least_squares.OvercontactLightCurveFit}
        return _cls[morphology]



import numpy as np

from copy import copy
from . orbit.container import OrbitalSupplements
from . import utils as bsutils
from .. import settings
from .. import (
    utils,
    const,
    umpy as up
)
from . curves.utils import compute_rel_d_geometry, compute_counterparts_rel_d_irrad
from .. base.types import INT, BOOL


def get_eclipse_boundaries(binary, components_distance):
    """
    Calculates the ranges in orbital azimuths (for phase=0 -> azimuth=pi/2)!!!  where eclipses occur.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param components_distance: float;
    :return: numpy.array;

    shape::

        [primary ecl_start, primary_ecl_stop, sec_ecl_start, sec_ecl_stop]
    """
    # check whether the inclination is high enough to enable eclipses
    if binary.morphology != 'over-contact':
        radius1 = np.mean([binary.primary.side_radius, binary.primary.forward_radius, binary.primary.backward_radius,
                           binary.primary.polar_radius])
        radius2 = np.mean([binary.secondary.side_radius, binary.secondary.forward_radius,
                           binary.secondary.backward_radius, binary.secondary.polar_radius])
        sin_i_critical = (radius1 + radius2) / components_distance
        sin_i = up.sin(binary.inclination)
        if sin_i < sin_i_critical:
            return np.array([const.HALF_PI, const.HALF_PI, const.PI, const.PI])
        radius1 = binary.primary.forward_radius
        radius2 = binary.secondary.forward_radius
        sin_i_critical = 1.01 * (radius1 + radius2) / components_distance
        square = up.power(sin_i_critical, 2) - up.power(up.cos(binary.inclination), 2)
        square = 0 if square < 0 else square
        square = 1 if square > 1 else square
        azimuth = up.arcsin(up.sqrt(square))
        azimuths = np.array([const.HALF_PI - azimuth, const.HALF_PI + azimuth, 1.5 * const.PI - azimuth,
                             1.5 * const.PI + azimuth]) % const.FULL_ARC
        return azimuths
    else:
        return np.array([0, const.PI, const.PI, const.FULL_ARC])


def find_apsidally_corresponding_positions(binary, radii, base_arr, supplement_arr, as_empty=None):
    """
    Function is intended to look for the couples of orbital positions from both sides of apsidal line that are most
    similar in terms of surface geometry (eg.: forward radius \approx forward radius).

    :param binary: elisa.binary_system.system.BinarySystem;
    :param radii: numpy.array; forward_radii
    :param base_arr: numpy.array; base orbital positions
    :param supplement_arr: numpy.array; orbital position from the opposite side
    :param as_empty: numpy.array; e.g. [np.nan, np.nan] depends on shape of base_arr item
    :return: elisa.binary_system.container.OrbitalSupplements;
    """
    if as_empty is None:
        as_empty = np.empty(5) * np.nan

    mean_r = np.mean(radii, axis=1)
    bigger_comp = np.argmax(mean_r)
    r_body = radii[:, base_arr[:, 0].astype(INT)]
    r_supplement = radii[:, supplement_arr[:, 0].astype(INT)]

    # finding indices of supplements_array closest to the base_array by comparing radius of larger component
    ids_of_closest_reduced_values = utils.find_idx_of_nearest(r_body[bigger_comp], r_supplement[bigger_comp])

    # making sure that found orbital positions are close enough to satisfy tolerance
    rel_geometry = compute_rel_d_geometry(binary, r_body[:, ids_of_closest_reduced_values], r_supplement)
    rel_geometry = np.max(rel_geometry, axis=0)
    is_supplement_geom = rel_geometry < settings.MAX_D_FLUX

    # making sure that found orbital positions are close enough to satisfy tolerance in relative irradiation
    args = (binary, base_arr[ids_of_closest_reduced_values, 1], supplement_arr[:, 1])
    rel_irrad = compute_counterparts_rel_d_irrad(*args)
    rel_irrad = np.max(rel_irrad, axis=0)
    is_supplement_irrad = rel_irrad < settings.MAX_D_FLUX

    is_supplement = np.logical_and(is_supplement_geom, is_supplement_irrad)

    # crating array which crates valid orbital position couples
    twin_in_reduced = np.full(ids_of_closest_reduced_values.shape, -1, dtype=INT)
    twin_in_reduced[is_supplement] = ids_of_closest_reduced_values[is_supplement]

    supplements = OrbitalSupplements()

    for id_supplement, id_reduced in enumerate(twin_in_reduced):
        id_reduced: INT = id_reduced
        args = (supplement_arr[id_supplement], as_empty)
        if id_reduced > -1:
            args = (base_arr[id_reduced], supplement_arr[id_supplement])

        if not utils.is_empty(args):
            supplements.append(*args)

    # treating a case of unmatched item in base array
    base_all_ids = up.arange(0, len(base_arr))
    is_not_in = ~np.isin(base_all_ids, twin_in_reduced)

    for is_not_in_id in base_all_ids[is_not_in]:
        if base_arr[is_not_in_id] not in supplement_arr:
            supplements.append(*(base_arr[is_not_in_id], as_empty))

    return supplements


def resolve_object_geometry_update(has_spots, size, rel_d, max_allowed_difference=None):
    """
    Evaluate where on orbital position is necessary to fully update geometry.
    Evaluation depends on difference of relative radii between neighbouring orbital positions.

    :param has_spots: bool; define if system has spots
    :param size: int; number of orbital positions
    :param rel_d: numpy.array; parameter characterizing change in flux due to change in surface geometry
    :param max_allowed_difference: float; maximum allowed change in flux change estimation between orbital positions
    :return: numpy.array[bool];
    """
    return _resolve_geometry_update(has_spots=has_spots, size=size, rel_d=rel_d, resolve="object",
                                    max_allowed_difference=max_allowed_difference or settings.MAX_D_FLUX)


def resolve_spots_geometry_update(spots_longitudes, size, pulsations_tests,
                                  max_allowed_difference=None):
    """
    Evaluate where on orbital position is necessary to fully update geometry.
    Evaluation depends on difference of spots longitudes between neighboring orbital positions.

    :param spots_longitudes: numpy.array; array of spot longitudes of spots during orbital motion
    :param size: int; number of orbital positions
    :param pulsations_tests: Tuple[bool]; True if component contains pulsations
    :param max_allowed_difference: float; maximum allowed difference of spot position on neighbouring orbital positions
    :return: Tuple[numpy.array[bool], numpy.array[bool]]; geometry update arrays for primary and secondary array.
                                                          If value in array is True, component geometry has to be
                                                          recalculated at a given orbital position.
    """
    reducer = {}
    for component in settings.BINARY_COUNTERPARTS.keys():
        if pulsations_tests[component]:
            # in case of pulsations, the geometry is recalculated always
            reducer[component] = np.ones(size, dtype=BOOL)
            continue

        # longitude of all spots stored in array (longitudes of the first spot are enough)
        longitude_array = np.array(list(utils.nested_dict_values(spots_longitudes[component]))[0]) if \
            not utils.is_empty(spots_longitudes[component]) else np.array([])

        d_long = np.abs(longitude_array - np.roll(longitude_array, shift=1))[1:]
        # creating 2*n array due to compatibility with new geometry assessment based on change in spot longitude where
        # both components are evaluated at once
        d_long = np.row_stack((d_long, d_long))

        reducer[component] = _resolve_geometry_update(
            has_spots=True, size=size, rel_d=d_long, resolve="spot",
            max_allowed_difference=max_allowed_difference or settings.MAX_SPOT_D_LONGITUDE
        )

    return reducer['primary'], reducer['secondary']


def _resolve_geometry_update(has_spots, size, rel_d, max_allowed_difference, resolve="object"):
    """
    Evaluate where on orbital position is necessary to fully update geometry.

    :param has_spots: bool; define if system has spots
    :param size: int; number of orbital positions
    :param rel_d: numpy.array; parameter characterizing change in flux due to change in surface geometry
    :param max_allowed_difference: float; maximum allowed change in flux change estimation between orbital positions
    :param resolve: str; decision parameter whether resolved object on eccentric orbit or spots movement,
                         "object" or "spots"
    :return: numpy.array[bool];
    """
    if resolve not in ["object", "spot"]:
        raise ValueError("Invalid option for `resolve`, use `object` or `spot`")

    # in case of spots, the boundary points will cause problems if you want to use the same geometry
    if has_spots and resolve == "object":
        return np.ones(size, dtype=BOOL)
    elif utils.is_empty(rel_d) and resolve == "spot":
        # if `rel_d` is empty and resolve is equal to `spot` it means given component has no spots
        # and does require build only on first position
        arr = up.zeros(size, dtype=BOOL)
        arr[0] = True
        return arr

    require_new_geo = np.ones(size, dtype=BOOL)

    cumulative_sum = np.array([0.0, 0.0])
    for i in range(1, size):
        cumulative_sum += rel_d[:, i - 1]
        if (cumulative_sum <= max_allowed_difference).all():
            require_new_geo[i] = False
        else:
            require_new_geo[i] = True
            cumulative_sum = np.array([0.0, 0.0])

    return require_new_geo


def resolve_irrad_update(rel_d_irrad, size):
    """
    Evaluate where new temperature distribution should be calculated

    :param rel_d_irrad: numpy.array; change in flux due to the change in mutual irradiation
    :param size: int; number of orbital positions
    :return: numpy.array[bool]; if true orbital position has to be recalculated due to change in reflected flux
    """
    require_new_build = np.ones(size, dtype=BOOL)

    cumulative_sum = np.array([0.0, 0.0])
    for i in range(1, size):
        cumulative_sum += rel_d_irrad[:, i - 1]
        if (cumulative_sum <= settings.MAX_D_FLUX).all():
            require_new_build[i] = False
        else:
            require_new_build[i] = True
            cumulative_sum = np.array([0.0, 0.0])

    return require_new_build


def phase_crv_symmetry(binary_system, phase: np.ndarray):
    """
    Utilizing symmetry of circular systems without spots and pulsations where you need to evaluate only half
    of the phases. Function finds such redundant phases and returns only unique phases.
    Expects phases from 0 to 1.0.

    :param binary_system: elisa.binary_system.system.BinarySystem;
    :param phase: numpy.array;
    :return: Tuple[numpy.array, numpy.array];
    """
    # keep those fucking methods imutable
    phase = phase.copy()
    if (not binary_system.has_pulsations()) & (not binary_system.has_spots()):
        symmetrical_counterpart = phase > 0.5
        phase[symmetrical_counterpart] = np.round(1.0 - phase[symmetrical_counterpart], 9)
        res_phases, reverse_idx = np.unique(phase, return_inverse=True)
        return res_phases, reverse_idx
    else:
        return phase, up.arange(phase.shape[0])


def in_eclipse_test(azimuths, ecl_boundaries):
    """
    Test whether in given phases eclipse occurs or not.
    !It works only for circular oribts!

    :param azimuths: Union[List, numpy.array];
    :param ecl_boundaries: Union[List, numpy.array];
    :return: numpy.array[bool];
    """

    if utils.is_empty(ecl_boundaries):
        return np.ones(len(azimuths), dtype=bool)

    if ecl_boundaries[0] < 1.5 * const.PI:
        primary_ecl_test = up.logical_and((azimuths >= ecl_boundaries[0]), (azimuths <= ecl_boundaries[1]))
    else:
        primary_ecl_test = up.logical_or((azimuths >= ecl_boundaries[0]), (azimuths < ecl_boundaries[1]))

    if ecl_boundaries[2] > const.HALF_PI:
        if ecl_boundaries[3] > const.HALF_PI:
            secondary_ecl_test = up.logical_and((azimuths >= ecl_boundaries[2]), (azimuths <= ecl_boundaries[3]))
        else:
            secondary_ecl_test = up.logical_or((azimuths >= ecl_boundaries[2]), (azimuths <= ecl_boundaries[3]))
    else:
        secondary_ecl_test = up.logical_and((azimuths >= ecl_boundaries[2]), (azimuths <= ecl_boundaries[3]))

    return up.logical_or(primary_ecl_test, secondary_ecl_test)


def correct_spot_positions_for_libration(system, phases):
    """
    Function corrects the position of spots for the libration motion caused by an eccentric orbit.

    :param system: Union[elisa.binary_system.system.BinarySystem,
                   elisa.binary_system.container.OrbitalPositionContainer];
    :param phases: numpy.array;
    :return: numpy.array; angular correction for each phase
    """
    # ensuring that phase = 0 is in dataset
    phases_p = [phases, ] if np.isscalar(phases) else copy(phases)
    phases_p = np.concatenate((phases_p, [0, ]))

    positions = system.calculate_orbital_motion(phases_p, return_nparray=True)

    ecc_anomaly = system.orbit.true_anomaly_to_eccentric_anomaly(positions[:, 3])
    mean_anomaly = system.orbit.eccentric_anomaly_to_mean_anomaly(ecc_anomaly)

    diff = mean_anomaly - positions[:, 3]
    diff = diff[:-1] - diff[-1]
    return diff if diff.shape[0] > 1 else diff[0]


def calculate_spot_longitudes(system, phases, component="all", correct_libration=True):
    """
    Function calculates the latitudes of every spot on given component(s) for every phase.

    :param system: Union[elisa.binary_system.system.BinarySystem,
                   elisa.binary_system.container.OrbitalPositionContainer];
    :param phases: numpy.array;
    :param component: str; 'primary' or 'secondary', if None both will be calculated
    :param correct_libration: bool; switch for calculation of correction for the libration motion of spots for EBs with
                                    eccentric orbit
    :return: Dict; {component: {spot_idx: np.array([....]), ...}, ...}
    """
    components = bsutils.component_to_list(component)
    components = {comp: getattr(system, comp) for comp in components}

    libration_correction = correct_spot_positions_for_libration(system, phases) if correct_libration else 0

    spots_longitudes = {
        comp: {
            spot_index: (instance.synchronicity - 1.0) * phases * const.FULL_ARC + spot.longitude + libration_correction
            for spot_index, spot in instance.spots.items()}
        for comp, instance in components.items()
    }
    return spots_longitudes


def assign_spot_longitudes(system, spots_longitudes, index=None, component="all"):
    """
    function assigns spot latitudes for each spot according to values in `spots_longitudes` in index `index`

    :param system: Union[elisa.binary_system.system.BinarySystem,
                   elisa.binary_system.container.OrbitalPositionContainer]
    :param spots_longitudes: Dict; {component: {spot_idx: np.array([....]), ...}, ...}, takes output of function
                                   `calculate_spot_latitudes`
    :param index: int; index of spot longitude values to be used, if none is given, scalar values are expected in
                      `spots_longitudes`
    :param component: str; 'primary' or 'secondary', if None both will be calculated
    """
    components = bsutils.component_to_list(component)
    components = {comp: getattr(system, comp) for comp in components}
    for comp, instance in components.items():
        for spot_index, spot in instance.spots.items():
            spot.longitude = spots_longitudes[comp][spot_index] if index is None else \
                spots_longitudes[comp][spot_index][index]



import numpy as np

from .. import utils as butils
from ... import atm, const
from ... import settings
from ... observer.passband import init_bolometric_passband
from ... binary_system import radius as bsradius
from ... base.curves.utils import get_component_limbdarkening_cfs


def get_limbdarkening_cfs(system, component="all", **kwargs):
    """
    Returns limb darkening coefficients for each face of each component.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param component: str;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float

    :return: Dict[str, numpy.array];
    """
    components = butils.component_to_list(component)

    symmetry_test = not system.has_spots() and not system.has_pulsations()

    return {c_name: get_component_limbdarkening_cfs(
        getattr(system, c_name),
        symmetry_test,
        kwargs['passband']
    ) for c_name in components}


def _get_normal_radiance(system, component="all", **kwargs):
    """
    Compute normal radiance for all faces and all components
    in elisa.binary_system.container.OrbitalPositionContainer.

    :param component: str;
    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param kwargs: Dict; arguments to be passed into light curve generator functions
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float

    :return: Dict[String, dict]
    """
    components = butils.component_to_list(component)
    symmetry_test = {
        component: not getattr(system, component).has_spots() and not getattr(system, component).has_pulsations()
        for component in components
    }
    temperatures, log_g = dict(), dict()

    # utilizing surface symmetry in case of a clear surface
    for cmpnt in components:
        component_instance = getattr(system, cmpnt)
        if symmetry_test[cmpnt]:
            temperatures[cmpnt] = component_instance.symmetry_faces(component_instance.temperatures)
            log_g[cmpnt] = component_instance.symmetry_faces(component_instance.log_g)
        else:
            temperatures[cmpnt] = component_instance.temperatures
            log_g[cmpnt] = component_instance.log_g

    retval = {
        component:
            atm.NaiveInterpolatedAtm.radiance(
                **dict(
                    temperature=temperatures[component],
                    log_g=log_g[component],
                    metallicity=getattr(system, component).metallicity,
                    atlas=getattr(getattr(system, component), "atmosphere") or settings.ATM_ATLAS,
                    **kwargs
                )
            ) for component in components
    }

    # mirroring symmetrical part back to the rest of the surface
    for cpmnt in components:
        if symmetry_test[cpmnt]:
            retval[cpmnt] = {fltr: getattr(system, cpmnt).mirror_face_values(vals) for
                             fltr, vals in retval[cpmnt].items()}

    return retval


def prep_surface_params(system, return_values=True, write_to_containers=False, **kwargs):
    """
    Prepares normal radiances and limb darkening coefficients variables.

    :param system: elisa.binary_system.container.OrbitalPositionContainer;
    :param return_values: bool; return normal radiances and limb darkening coefficients
    :param write_to_containers: bool; calculated values will be assigned to `system` container
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float
    :return: Tuple;
    """
    # obtain limb darkening factor for each face
    ld_cfs = get_limbdarkening_cfs(system, **kwargs)
    # compute normal radiance for each face and each component
    normal_radiance = _get_normal_radiance(system, **kwargs)

    # checking if `bolometric` filter is already used
    if 'bolometric' in ld_cfs['primary']:
        bol_ld_cfs = {
            component: {'bolometric': ld_cfs[component]['bolometric']}
            for component in settings.BINARY_COUNTERPARTS
        }
    else:
        passband, left_bandwidth, right_bandwidth = init_bolometric_passband()
        bol_kwargs = {
            'passband': {'bolometric': passband},
            'left_bandwidth': left_bandwidth,
            'right_bandwith': right_bandwidth,
            'atlas': 'whatever'
        }
        bol_ld_cfs = get_limbdarkening_cfs(system, **bol_kwargs)

    normal_radiance = atm.correct_normal_radiance_to_optical_depth(normal_radiance, bol_ld_cfs)

    if write_to_containers:
        for component in settings.BINARY_COUNTERPARTS:
            star = getattr(system, component)
            setattr(star, 'normal_radiance', normal_radiance[component])
            setattr(star, 'ld_cfs', ld_cfs[component])

    return normal_radiance, ld_cfs if return_values else None


def update_surface_params(require_rebuild, container, normal_radiance, ld_cfs, **kwargs):
    """
    Function either recalculates normal radiances and limb darkening coefficients or it assigns old values to the
    container according to the `require_rebuild` condition.

    :param require_rebuild: bool; testing condition for recalculation of surface parameters
    :param container: elisa.binary_system.container.OrbitalPositionContainer;
    :param normal_radiance: Dict; old values of normal radiances
    :param ld_cfs: Dict; old values of limb darkening coefficients
    :param kwargs: Dict;
    :return: Tuple; updated container and updated normal radiances and limb darkening coefficients
    """
    if require_rebuild:
        normal_radiance, ld_cfs = \
            prep_surface_params(container, return_values=True, write_to_containers=True, **kwargs)
    else:
        for component in settings.BINARY_COUNTERPARTS:
            star = getattr(container, component)
            setattr(star, 'normal_radiance', normal_radiance[component])
            setattr(star, 'ld_cfs', ld_cfs[component])

    return container, normal_radiance, ld_cfs


def split_orbit_by_apse_line(orbital_motion, orbital_mask):
    """
    Split orbital positions represented by `orbital_motion` array on two groups separated by line of apsides.
    Separation is defined by `orbital_mask`

    :param orbital_motion: numpy.array; arraywhcih represents orbital positions
    :param orbital_mask: numpy.array[bool]; mask which defines separation (True is one side and False is other side)
    :return: Tuple[numpy.array, numpy.array];
    """
    reduced_orbit_arr = orbital_motion[orbital_mask]
    supplement_to_reduced_arr = orbital_motion[~orbital_mask]
    return reduced_orbit_arr, supplement_to_reduced_arr


def forward_radii_from_distances(binary, distances, potentials=None):
    """
    Returns forward radii for each components distance.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param distances: numpy.array;
    :param potentials: numpy.array; if None, they will be calculated
    :return: numpy.array; 2*N array of forward radii for each component
    """
    corrected_potentials = binary.correct_potentials(distances=distances, component="all", iterations=2) \
        if potentials is None else potentials

    pargs = (distances, corrected_potentials['primary'], binary.mass_ratio, binary.primary.synchronicity, "primary")
    sargs = (distances, corrected_potentials['secondary'], binary.mass_ratio, binary.secondary.synchronicity,
             "secondary")

    return np.vstack((bsradius.calculate_forward_radii(*pargs), bsradius.calculate_forward_radii(*sargs)))


def compute_rel_d_geometry(binary, radii, radii_counterpart):
    """
    Function estimates the maximum change in flux due to the change in the geometry estimated by the change in the
    forward radius.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param radii: numpy.array;
    :param radii_counterpart: numpy.array;
    :return: numpy.array;
    """
    eq_radii = np.array([binary.primary.equivalent_radius, binary.secondary.equivalent_radius])
    fwd_r_diff = np.abs(radii_counterpart - radii)

    d_flux = 2 * eq_radii[:, np.newaxis] * fwd_r_diff + fwd_r_diff ** 2
    total_flux = eq_radii ** 2
    return d_flux / np.sum(total_flux)


def relative_irradiation(binary, distances):
    """
    Returns an estimate of how much of the component flux comes from the reflected light of the companion

    :param binary: elisa.binary_system.system.BinarySystem;
    :param distances: numpy.array; orbital distances
    :return: numpy.array;
    """
    temp_ratio4 = np.power(binary.primary.t_eff / binary.secondary.t_eff, 4)
    r_ratio2 = np.power(binary.primary.equivalent_radius / binary.secondary.equivalent_radius, 2)
    coeff = r_ratio2 * temp_ratio4
    irrad1 = np.power(binary.primary.equivalent_radius / distances, 2) / (1 + coeff)
    irrad2 = np.power(binary.secondary.equivalent_radius / distances, 2) / (1 + 1 / coeff)
    return np.vstack((irrad1, irrad2))


def compute_counterparts_rel_d_irrad(binary, distances, distances_counterpart):
    """
    Estimates a relative change in recieved irradiation from a companion.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param distances: numpy.array; orbital distances
    :param distances_counterpart: numpy.array; orbital distances
    :return: numpy.array;
    """
    irrad = relative_irradiation(binary, distances)
    irrad_c = relative_irradiation(binary, distances_counterpart)

    return np.abs(irrad - irrad_c)


def compute_rel_d_irradiation(binary, distances):
    """
    Estimates a relative change in recieved irradiation from a companion between the most similar orbital positions.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param distances: numpy.array; orbital distances (sorted)
    :return: numpy.array;
    """
    irrad = relative_irradiation(binary, distances)
    return np.abs(irrad[:, 1:] - irrad[:, :-1])


def compute_rel_d_radii_from_counterparts(radii, base_positions, mirrors):
    """
    Returns relative differences between forward radii between two orbital counterparts.

    :param radii: numpy.array; forward radii
    :param base_positions: numpy.array; base orbital position array
    :param mirrors: numpy.array; orbital counterposition array
    :return: np.array; (2 * N) of relative changes in relative distances
    """
    fwd_radii_base = radii[base_positions[:, 0]]
    fwd_radii_counterpart = radii[mirrors[:, 0]]
    return np.abs(fwd_radii_base - fwd_radii_counterpart) / fwd_radii_base.mean(axis=1)[:, np.newaxis]


def prepare_apsidaly_symmetric_orbit(binary, azimuths, phases):
    """
    Prepare set of orbital positions that are symmetrical in therms of surface geometry. For each couple, the orbital
    position is mirrored using apsidal line in order to reduce time for generating the light curve.

    :param binary: elisa.binary_star.system.BinarySystem;
    :param azimuths: numpy.array; orbital azimuths of positions in which LC will be calculated
    :param phases: numpy.array; orbital phase of positions in which LC will be calculated
    :return: Tuple;


    shape ::

        (numpy.array, list, numpy.array)

        - unique_phase_indices - numpy.array : indices that points to the orbital positions from one half of the
        orbital motion divided by apsidal line
        - orbital_motion_counterpart - list - Positions produced by mirroring orbital positions given by
        indices `unique_phase_indices`
        - orbital_motion_array_counterpart - numpy.array - sa as `orbital_motion_counterpart` but in numpy.array form
    """
    azimuth_boundaries = [binary.argument_of_periastron, (binary.argument_of_periastron + const.PI) % const.FULL_ARC]

    if azimuth_boundaries[0] < azimuth_boundaries[1]:
        unique_geometry = np.logical_and(azimuths >= azimuth_boundaries[0], azimuths < azimuth_boundaries[1])
    else:
        unique_geometry = np.logical_xor(azimuths <= azimuth_boundaries[0], azimuths > azimuth_boundaries[1])

    unique_phase_indices = np.arange(phases.shape[0])[unique_geometry]
    unique_geometry_azimuths = azimuths[unique_geometry]
    unique_geometry_counterazimuths = (2 * binary.argument_of_periastron - unique_geometry_azimuths) % const.FULL_ARC

    kwargs = dict(input_argument=unique_geometry_counterazimuths, return_nparray=True, calculate_from='azimuth')
    orbital_motion_array_counterpart = binary.calculate_orbital_motion(**kwargs)

    return unique_phase_indices, orbital_motion_array_counterpart, unique_geometry


def adjust_eclipse_width(true_anomalies, true_anomaly_of_eclipse):
    """
    Extends the angular width of the eclipse by the separation of the true anomalies near the eclipse to smooth out 
    the transition before/after the eclipse.
    
    :param true_anomalies: numpy.array; true anomalies of the orbital positions
    :param true_anomaly_of_eclipse: float; true anomaly of the eclipse
    :return: float; adjusted width of an eclipse
    """
    distances = np.abs(true_anomalies - true_anomaly_of_eclipse)
    inverse_points_mask = distances > const.PI
    distances[inverse_points_mask] = const.FULL_ARC - distances[inverse_points_mask]
    idxs = np.argsort(distances)[:2]
    return 1.5 * np.abs(true_anomalies[idxs[1]] - true_anomalies[idxs[0]])



import numpy as np

from .. types import INT
from ... import settings, ld, umpy as up
from ... observer.passband import init_rv_passband


def include_passband_data_to_kwargs(**kwargs):
    """
    Including dummy passband from which radiometric radial velocities will be calculated.

    :param kwargs: Tuple;
    :return: Tuple;
    """
    psbnd, right_bandwidth, left_bandwidth = init_rv_passband()
    kwargs.update({
        'passband': {'rv_band': psbnd},
        'left_bandwidth': left_bandwidth,
        'right_bandwidth': right_bandwidth
    })
    return kwargs


def calculate_surface_element_fluxes(band, star):
    """
    Function generates outgoing flux from each surface element of given star container in certain band.

    :param star: elisa.base.container.StarContainer; star container with all necessary parameters pre-calculated
    :param band: str; name of the photometric band compatible with supported names in config
    :return: numpy.array
    """
    indices = star.indices
    radiance = star.normal_radiance[band][indices]
    ld_cfs = star.ld_cfs[band][indices]
    cosines = star.los_cosines[indices]
    coverage = star.coverage[indices]

    ld_cors = ld.limb_darkening_factor(coefficients=ld_cfs,
                                       limb_darkening_law=settings.LIMB_DARKENING_LAW,
                                       cos_theta=cosines)

    return radiance * cosines * coverage * ld_cors


def flux_from_star_container(band, star):
    """
    Function generates outgoing flux from given star container in certain band.

    :param star: elisa.base.container.StarContainer; star container with all necessary parameters pre-calculated
    :param band: str; name of the photometric band compatible with supported names in config,
    :return: numpy.array;
    """

    return up.sum(calculate_surface_element_fluxes(band, star))


def generate_teff_logg_for_ld_cfs(component_instance, symmetry_test):
    """
    Generates temperatures and log_g parameters either for full star or symmetrical part based on the symmetry test.

    :param component_instance: elisa.base.container.StarContainer; star container
    :param symmetry_test: bool; whether to use star symmetry
    :return: Tuple;
    """
    if settings.USE_SINGLE_LD_COEFFICIENTS:
        temperatures = np.array([component_instance.t_eff, ])
        log_g = np.array([np.max(component_instance.log_g), ])
    elif symmetry_test:
        temperatures = component_instance.symmetry_faces(component_instance.temperatures)
        log_g = component_instance.symmetry_faces(component_instance.log_g)
    else:
        temperatures = component_instance.temperatures
        log_g = component_instance.log_g

    return temperatures, log_g


def get_component_limbdarkening_cfs(component_instance, symmetry_test, passbands):
    """
    Returns limb darkening coefficients for given component in each passband.

    :param component_instance: elisa.base.container.StarContainer;
    :param symmetry_test: bool;
    :param passbands: list;
    :return: Dict;
    """
    if component_instance.limb_darkening_coefficients is not None:
        desired_repeats = (component_instance.temperatures.shape[0], 1)
        try:
            ld_cfs = {passband: np.tile(component_instance.limb_darkening_coefficients[passband], desired_repeats)
                      for passband in passbands}
        except KeyError:
            missing_passbands = list(set(passbands) - set(component_instance.limb_darkening_coefficients.keys()))
            raise KeyError(f'Please supply limb-darkening factors for {missing_passbands} pasband(s) as well.')
    else:
        temperatures, log_g = generate_teff_logg_for_ld_cfs(component_instance, symmetry_test)

        ld_cfs = ld.interpolate_on_ld_grid(
            temperature=temperatures,
            log_g=log_g,
            metallicity=component_instance.metallicity,
            passband=passbands
        )

        if symmetry_test:
            if settings.USE_SINGLE_LD_COEFFICIENTS:
                ld_cfs = {fltr: vals[np.zeros(component_instance.temperatures.shape, dtype=INT)]
                          for fltr, vals in ld_cfs.items()}
            else:
                ld_cfs = {fltr: component_instance.mirror_face_values(vals)
                          for fltr, vals in ld_cfs.items()}

    return ld_cfs



import numpy as np

from typing import Dict
from copy import copy
from abc import (
    ABCMeta,
    abstractmethod
)

from .. utils import is_empty
from .. base.spot import Spot
from .. logger import getLogger
from .. import (
    umpy as up
)
from .. import settings

logger = getLogger('base.body')


class Body(metaclass=ABCMeta):
    """
    Abstract class that defines bodies modelled by this package.
    """

    ID = 1
    MANDATORY_KWARGS = []
    OPTIONAL_KWARGS = []
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS

    def __init__(self, name: str, **kwargs):
        """
        Properties of abstract class Body.
        """
        # initial kwargs
        self.kwargs: Dict = copy(kwargs)

        if is_empty(name):
            self.name = str(Body.ID)
            logger.debug(f"name of class instance {self.__class__.__name__} set to {self.name}")
            Body.ID += 1
        else:
            self.name = str(name)

        # initializing paramas to default values
        self.synchronicity: float = np.nan
        self.mass: float = np.nan
        self.albedo: float = np.nan
        self.discretization_factor: float = np.float64(up.radians(settings.DEFAULT_DISCRETIZATION_FACTOR))
        self.t_eff: float = np.nan
        self.polar_radius: float = np.nan
        self._spots: Dict = dict()
        self.equatorial_radius: float = np.nan
        self.atmosphere: str = ""

    @abstractmethod
    def init(self):
        pass

    @abstractmethod
    def transform_input(self, *args, **kwargs):
        pass

    @property
    def spots(self):
        """
        :return: Dict[int, elisa.base.spot.Spot]
        """
        return self._spots

    @spots.setter
    def spots(self, spots):
        """
        Order in which the spots are defined will determine the layering of the spots (spot defined as first will lay
        bellow any subsequently defined overlapping spot). Example of defined spots

        ::

            [
                 {"longitude": 90,
                  "latitude": 58,
                  "angular_radius": 15,
                  "temperature_factor": 0.9},
                 {"longitude": 85,
                  "latitude": 80,
                  "angular_radius": 30,
                  "temperature_factor": 1.05},
                 {"longitude": 45,
                  "latitude": 90,
                  "angular_radius": 30,
                  "temperature_factor": 0.95},
             ]

        :param spots: Iterable[Dict]; definition of spots for given object
        """
        self._spots = {idx: Spot(**spot_meta) for idx, spot_meta in enumerate(spots)} if not is_empty(spots) else dict()
        for spot_idx, spot_instance in self.spots.items():
            self.setup_spot_instance_discretization_factor(spot_instance, spot_idx)

    def has_spots(self):
        """
        Find whether object has defined spots.

        :return: bool;
        """
        return len(self._spots) > 0

    def remove_spot(self, spot_index: int):
        """
        Remove n-th spot index of object.

        :param spot_index: int;
        """
        del self._spots[spot_index]

    def setup_spot_instance_discretization_factor(self, spot_instance, spot_index):
        """
        Setup discretization factor for given spot instance based on defined rules.

        - use value of the parent star if the spot discretization factor is not defined
        - if spot_instance.discretization_factor > 0.5 * spot_instance.angular_diameter then factor is set to
                      0.5 * spot_instance.angular_diameter

        :param spot_instance: elisa.base.spot.Spot;
        :param spot_index: int; spot index (has no affect on process, used for logging)
        :return: elisa.base.spot.Spot;
        """
        if is_empty(spot_instance.discretization_factor):
            logger.debug(f'angular density of the spot {spot_index} on {self.name} component was not supplied '
                         f'and discretization factor of star {self.discretization_factor} was used.')
            spot_instance.discretization_factor = 0.9 * self.discretization_factor
        if spot_instance.discretization_factor > spot_instance.angular_radius:
            logger.debug(f'angular density {self.discretization_factor} of the spot {spot_index} on {self.name} '
                         f'component was larger than its angular radius. Therefore value of angular density was '
                         f'set to be equal to 0.5 * angular diameter')
            spot_instance.discretization_factor = spot_instance.angular_radius

        return spot_instance



from ... import umpy as up

from ... base.curves import rv_point
from . import c_router


def com_radial_velocity(single, **kwargs):
    """
    Calculates radial velocity curves of the `single` system using radial velocity of centres of masses.

    :param single: elisa.single_system.SingleSystem
    :return:
    """
    phases = kwargs.pop("phases")
    return {'star': single.gamma * up.ones(phases.shape[0])}


def compute_rv_curve_without_pulsations(single, **kwargs):
    """
    Function for calculation radial velocity curves with radiometric methods for single system without pulsations.

    :param single: elisa.single_system.system.SingleSystem;
    :param kwargs: Dict;
    :return: Dict[str, numpy.array];
    """
    initial_system = c_router.prep_initial_system(single)
    rv_labels = ['star', ]
    args = single, initial_system, kwargs.pop("phases"), rv_point.compute_rv_at_pos, rv_labels
    return c_router.produce_curves_wo_pulsations(*args, **kwargs)


def compute_rv_curve_with_pulsations(single, **kwargs):
    """
    Function for calculation radial velocity curves with radiometric methods for single system without pulsations.

    :param single: elisa.single_system.system.SingleSystem;
    :param kwargs: Dict;
    :return: Dict[str, numpy.array];
    """
    initial_system = c_router.prep_initial_system(single, **dict(build_pulsations=False))
    rv_labels = ['star', ]
    args = single, initial_system, kwargs.pop("phases"), rv_point.compute_rv_at_pos, rv_labels
    return c_router.produce_curves_with_pulsations(*args, **kwargs)



import numpy as np

from .. import utils as bsutils
from ... pulse.container_ops import (
    incorporate_pulsations_to_model,
    generate_harmonics
)
from ... import const


def build_harmonics(system, component, components_distance):
    """
    Adds pre-calculated spherical harmonics values for each pulsation mode.

    :param system: elisa.binary_system.contaier.OrbitalPositionContainer; instance
    :param component: Union[str, None];
    :param components_distance: float;
    """
    components = bsutils.component_to_list(component)
    for component in components:
        star = getattr(system, component)
        pos_correction = bsutils.correction_to_com(system.position.distance, system.mass_ratio, system.secondary.com)[0]
        asini = system.semi_major_axis * np.sin(system.inclination)
        if star.has_pulsations():
            phase = bsutils.calculate_rotational_phase(system, component)
            com_x = 0 if component == 'primary' else components_distance
            # LTE effect
            time_correction = (star.com[0] - pos_correction) * asini / const.C
            generate_harmonics(star, com_x=com_x, phase=phase, time=system.time+time_correction)


def build_perturbations(system, component, components_distance):
    """
    Incorporating perturbations of surface quantities into the PositionContainer.

    :param system: elisa.binary_system.contaier.OrbitalPositionContainer; instance
    :param component: Union[str, None];
    :param components_distance: float;
    :return: elisa.binary_system.contaier.OrbitalPositionContainer; instance
    """
    components = bsutils.component_to_list(component)
    for component in components:
        star = getattr(system, component)
        if star.has_pulsations():
            com_x = 0 if component == 'primary' else components_distance
            incorporate_pulsations_to_model(star, com_x=com_x, scale=system.semi_major_axis)
    return system



import os
import sys
from typing import Tuple

import numpy as np
import pandas as pd

from scipy import interpolate

from . plot import PassbandPlot
from .. import settings
from ..base.types import FLOAT


def init_bolometric_passband():
    """
    initializing bolometric passband and its wavelength boundaries

    :return: Tuple;
    """
    df = pd.DataFrame(
        {
            settings.PASSBAND_DATAFRAME_THROUGHPUT: [1.0, 1.0],
            settings.PASSBAND_DATAFRAME_WAVE: [50.0, 2000000.0]
        }
    )
    right_bandwidth = sys.float_info.max
    left_bandwidth = 0.0
    bol_passband = PassbandContainer(table=df, passband='bolometric')

    return bol_passband, right_bandwidth, left_bandwidth


def init_rv_passband():
    """
    Initializing passband used to calculate radial velocities

    :return: Tuple
    """
    df = pd.DataFrame(
        {settings.PASSBAND_DATAFRAME_THROUGHPUT: [1.0, 1.0],
         settings.PASSBAND_DATAFRAME_WAVE: settings.RV_LAMBDA_INTERVAL})
    right_bandwidth = settings.RV_LAMBDA_INTERVAL[1]
    left_bandwidth = settings.RV_LAMBDA_INTERVAL[0]
    psmbnd = PassbandContainer(table=df, passband='rv_band')

    return psmbnd, right_bandwidth, left_bandwidth


def bolometric(x):
    """
    Bolometric passband interpolation function in way of lambda x: 1.0

    :param x:
    :return: float or numpy.array; 1.0s in shape of x
    """
    if isinstance(x, (float, int)):
        return 1.0
    if isinstance(x, list):
        return [1.0] * len(x)
    if isinstance(x, np.ndarray):
        return np.array([1.0] * len(x))


class PassbandContainer(object):
    def __init__(self, table, passband):
        """
        Data container used for storing passband response curves. Fully initialized PassbandContainers contain following
        attributes:

            - left_bandwidth, right_bandwidth: left and right wavelength boundary of the passband
            - table: pandas.DataFrame; dataframe containing a `wavelength` column with corresponding `throughput` values
                                       defining a given passband
            - passband: name of the passband

        The response curve is stored in a pandas.DataFrame
        Setup PassbandContainier object. It carres dependedncies of throughputs on wavelengths for given passband.

        :param table: pandas.DataFrame;
        :param passband: str;
        """
        self.left_bandwidth = np.nan
        self.right_bandwidth = np.nan
        self.akima = None
        self._table = pd.DataFrame({})
        self.wave_unit = "angstrom"
        self.passband = passband
        # in case this np.pi will stay here, there will be rendundant multiplication in intensity integration
        self.wave_to_si_mult = 1e-10
        self.plot = PassbandPlot(self)

        setattr(self, 'table', table)

    @property
    def table(self):
        """
        Return pandas dataframe which represent pasband table as dependecy of throughput on wavelength.

        :return: pandas.DataFrame;
        """
        return self._table

    @table.setter
    def table(self, df):
        """
        Setter for passband table.
        It precomputes left and right bandwidth for given table and also interpolation function placeholder.
        Akima1DInterpolator is used. If `bolometric` passband is used then interpolation function is like::

            lambda x: 1.0


        :param df: pandas.DataFrame;
        """
        self._table = df
        self.akima = bolometric if (self.passband.lower() in ['bolometric', 'rv_band']) else \
            interpolate.Akima1DInterpolator(df[settings.PASSBAND_DATAFRAME_WAVE],
                                            df[settings.PASSBAND_DATAFRAME_THROUGHPUT])
        self.left_bandwidth = df[settings.PASSBAND_DATAFRAME_WAVE].min()
        self.right_bandwidth = df[settings.PASSBAND_DATAFRAME_WAVE].max()

    @staticmethod
    def get_passband_df(passband):
        """
        Read content of passband table (csv file) based on passband name.

        :param passband: str;
        :return: pandas.DataFrame;
        """
        if passband not in settings.PASSBANDS:
            raise ValueError('Invalid or unsupported passband function.')
        file_path = os.path.join(settings.PASSBAND_TABLES, str(passband) + '.csv')
        df = pd.read_csv(file_path)
        df[settings.PASSBAND_DATAFRAME_WAVE] *= 10.0
        return df

    def get_bandwidth(self) -> Tuple[FLOAT, FLOAT]:
        return self.left_bandwidth, self.right_bandwidth

    @classmethod
    def from_name(cls, passband: str):
        df = cls.get_passband_df(passband)
        return cls(table=df, passband=passband)



import numpy as np
from copy import copy

from jsonschema import (
    validate,
    ValidationError
)

from .. base.error import YouHaveNoIdeaError
from .. import const, utils, settings, units
from .. base.transform import StarProperties, SystemProperties


def move_sys_onpos(system, position, on_copy=True):
    """
    Prepares a postion container for given orbital position.
    Supplied `system` is not affected if `on_copy` is set to True.

    Following methods are applied::

        system.set_on_position_params()
        system.flat_it()
        system.apply_rotation()
        system.apply_darkside_filter()

    :param position: collections.namedtuple; elisa.const.Position;
    :param system: elisa.single_system.container.PositionContainer;
    :param on_copy: bool;
    :return: container; elisa.sinary_system.container.PositionContainer;
    """
    if on_copy:
        system = system.copy()
    system.set_on_position_params(position)
    system.flat_it()
    system.apply_rotation()
    system.add_secular_velocity()
    system.calculate_face_angles(line_of_sight=const.LINE_OF_SIGHT)
    system.apply_darkside_filter()
    return system


def calculate_volume(system):
    """
    Returns volume of rotationally squashed star based on a volume of elipsoid.

    :param system: elisa.SingleSystem;
    :return: float
    """
    args = system.star.polar_radius, system.star.equatorial_radius, system.star.equatorial_radius
    return utils.calculate_ellipsoid_volume(*args)


def validate_single_json(data):
    """
    Validate input json to create SingleSystem instance from.

    :param data: Dict; json like object
    :return: ; return True if valid schema, otherwise raise error
    :raise: ValidationError;
    """
    schema_std = settings.SCHEMA_REGISTRY.get_schema("single_system_std")
    schema_radius = settings.SCHEMA_REGISTRY.get_schema("single_system_radius")
    std_valid, radius_valid = False, False

    try:
        validate(instance=data, schema=schema_std)
        std_valid = True
    except ValidationError:
        pass

    try:
        validate(instance=data, schema=schema_radius)
        radius_valid = True
    except ValidationError:
        pass

    if not std_valid and not radius_valid:
        raise YouHaveNoIdeaError("Make sure that list of parameters is consistent with the used schema.")

    if radius_valid & std_valid:
        raise YouHaveNoIdeaError("Make sure that list of fitted parameters contain only `standard` or `radius` "
                                 "combination of parameter (containing either `polar_log_g` or `polar_radius`).")

    return True


def resolve_json_kind(data):
    """
    Resolve if json is in `standard` or `radius` format.

    std - size of the star defined by the polar surface gravity `polar_log_g` parameter
    community - size of the star defined by the `equivalent_radius` parameter

    :param data: Dict; json like
    :return: str; `std` or `radius`
    """
    polar_g = data['star'].get('polar_log_g')
    polar_radius = data['star'].get('equivalent_radius')

    if polar_g:
        return "std"
    elif polar_radius:
        return "radius"
    raise LookupError("It seems your JSON is invalid.")


def transform_json_radius_to_std(data):
    """
    Transform `radius` input format json to `std` json.
    Compute polar_log_g form equivalent radius.

    :param data: Dict;
    :return: Dict;
    """
    def equatorial_to_polar_radius(r_eq, period, mass):
        k = 2 * np.power(const.PI, 2) * np.power(r_eq, 3) / (const.G * mass * np.power(period, 2))
        return 1 / (1 - k)

    def polar_from_equatorial_radius(r_eq, period, mass):
        rho = equatorial_to_polar_radius(r_eq, period, mass)
        return r_eq / np.power(rho, 2.0/3.0)

    mass = StarProperties.mass(data['star']['mass'])
    # default unit of radius is the same as for the semi-major axis
    radius = SystemProperties.semi_major_axis(data['star'].pop('equivalent_radius'))

    period_val = copy(data["system"]["rotation_period"])
    period = (SystemProperties.period(period_val) * units.DefaultSingleSystemUnits.system.rotation_period).\
        to(units.TIME_UNIT).value

    polar_radius = polar_from_equatorial_radius(radius, period, mass)
    data['star']['polar_log_g'] = np.log10(const.G * mass / np.power(polar_radius, 2)) + 2.0

    return data



import os
import json
import numpy as np
import os.path as op
from time import time

from datetime import datetime
from typing import Dict

from .. params import parameters
from ... import settings
from ... logger import getPersistentLogger


logger = getPersistentLogger('analytics.binary_fit.mixins')


class MCMCMixin(object):
    """
    Module for handling of the MCMC chain and the sampler.
    """
    @staticmethod
    def renormalize_flat_chain(flat_chain, all_lables, labels, normalization):
        """
        Re-normalize values in chain stored within (0, 1) interval to their original values.

        :param flat_chain: numpy.array; resulting flatten chain obtained from MCMC sampling
        :param all_lables: List; names of all variable model parameters
        :param labels: List; names variable model parameters desired in the output array
        :param normalization: Dict[str, tuple]; normalization bounds for variable model parameters
        :return: numpy.array; re-normalized flat chain
        """
        retval = []
        for ii, label in enumerate(labels):
            idx = all_lables.index(label)
            retval.append(parameters.renormalize_value(flat_chain[:, idx],
                                                       normalization[label][0], normalization[label][1]))

        retval = np.column_stack(retval)
        return retval

    @staticmethod
    def resolve_mcmc_result(flat_chain, fitable, normalization, percentiles=None):
        """
        Function process flat chain (output from McMcFit._fit.get_chain(flat=True)) and produces dictionary with
        results.

        :param normalization: Dict; normalisation map
        :param flat_chain: emcee.ensemble.EnsembleSampler.get_chain(flat=True);
        :param fitable: Dict; fitable parameters
        :param percentiles: List; [percentile for left side error estimation, percentile of the centre,
                                   percentile for right side error estimation]
        :return: Dict; JSON with variable model parameters in flat format
        """
        percentiles = [16, 50, 84] if percentiles is None else percentiles
        result = dict()
        for idx, key in enumerate(fitable):
            mcmc_result = np.percentile(flat_chain[:, idx], percentiles)
            vals = parameters.renormalize_value(mcmc_result, normalization[key][0], normalization[key][1])

            # rounding up values to significant digits
            sigma = np.min(np.abs([vals[2] - vals[1], vals[1] - vals[0]]))
            prec = - int(np.log10(sigma)) + 1
            vals = np.round(vals, decimals=prec)

            result[key] = {
                "value": vals[1],
                "confidence_interval": {
                    "min": min(vals),
                    "max": max(vals)},
                "fixed": False,
                "min": normalization[key][0],
                "max": normalization[key][1],
                "unit": fitable[key].to_dict()['unit']
            }

        return result

    @staticmethod
    def save_flat_chain(flat_chain: np.array, fitable: Dict, norm: Dict, fit_id=None):
        """
        Store samples of the MCMC run.

        :param fit_id: str; id or location (ending with .json) which identifies fit file (if not specified, current
                            datetime is used)
        :param flat_chain: numpy.array; flatted array of parameters values in each mcmc step::

            [[param0_0, param1_0, ..., paramk_0],
            [param0_1, param1_1, ..., paramk_1], ...
            [param0_b, param1_n, ..., paramk_n]]

        :param norm: Dict; normalization dict
        :param fitable: Union[List, numpy.array]; labels of parameters in order of params in `flat_chain`
        """
        home = settings.HOME
        if fit_id is not None:
            if op.isdir(op.dirname(fit_id)):
                fdir = op.dirname(fit_id)
                fname = op.basename(fit_id)
                home = ''
            else:
                fdir = str(fit_id)
                fname = f'{str(fit_id)}.json'
        else:
            now = datetime.now()
            fdir = now.strftime(settings.DATE_MASK)
            fname = f'{now.strftime(settings.DATETIME_MASK)}.json'

        fpath = op.join(home, fdir, fname)
        os.makedirs(op.join(settings.HOME, fdir), exist_ok=True)
        data = {
            "flat_chain": flat_chain.tolist() if isinstance(flat_chain, np.ndarray) else flat_chain,
            "fitable_parameters": list(fitable.keys()),
            "normalization": norm,
            "fitable": {key: val.to_dict() for key, val in fitable.items()}
        }
        with open(fpath, "w") as f:
            f.write(json.dumps(data, indent=4))
        logger.info(f'MCMC chain, variable`s fitable and normalization constants were stored in: {fpath}')
        return fpath

    @staticmethod
    def load_flat_chain(fit_id):
        """
        Lead the result (flat chain) from the MCMC run.

        :param fit_id: str; id or location (ending with .json) which identifies fit file (if not specified, current
                            datetime is used)
        :return: Dict;  flat_chain, variable parameters, normalization bounds, JSON with results
        """
        fname = fit_id if str(fit_id).endswith('.json') else f'{fit_id}.json'

        # expected full path
        if op.isfile(fname):
            fpath = fname
        else:
            # expect timestamp default name
            fdir = fit_id[:len(settings.DATE_MASK) + 2]
            fpath = op.join(settings.HOME, fdir, fname)
            if not os.path.isfile(fpath):
                # expected user defined fit_id
                fpath = op.join(settings.HOME, fit_id, fname)

        with open(fpath, "r") as f:
            return json.loads(f.read())

    @staticmethod
    def worker(sampler, p0, nsteps, nsteps_burn_in,
               save=False, fit_id=None, fitable=None, normalization=None, progress=False):
        """
        Multiprocessor worker for MCMC sampling routine.

        :param sampler: emcee.EnsembleSampler;
        :param p0: numpy.array; (n_walkers x n_variables) distribution of normalized parameters in the initial step
        :param nsteps: int; number of MCMC sampling iterations
        :param nsteps_burn_in: int; initial discarded iterations of the MCMC corresponding to the expected
                                    thermalization portion of the chain
        :param save: bool; if true, the MCMC flat chain will be stored
        :param fit_id: str; id or location (ending with .json) which identifies fit file (if not specified, current
                            dateime is used)
        :param fitable: Dict; JSON containing variable model parameters
        :param normalization: Dict[str, tuple]; normalization boundaries
        :param progress: bool; display the progress bar of the sampling
        """
        logger.info("running burn-in...")
        p0, _, _ = sampler.run_mcmc(p0, nsteps_burn_in, progress=progress, store=False) \
                       if nsteps_burn_in > 0 else p0, None, None
        sampler.reset()
        logger.info("running production...")

        if save:
            t_between_dumps = time()
            for _ in sampler.sample(p0, iterations=nsteps, progress=progress):
                if time() - t_between_dumps > settings.MCMC_SAVE_INTERVAL:
                    MCMCMixin.save_flat_chain(sampler.get_chain(flat=True), fitable=fitable, norm=normalization,
                                              fit_id=fit_id)
                    t_between_dumps = time()
        else:
            _, _, _ = sampler.run_mcmc(p0, nsteps, progress=progress)

    @staticmethod
    def mcmc_nwalkers_vs_ndim_validity_check(nwalkers, ndim):
        """
        Validate number of MCMC walkers satisfies the condition `nwalkers < ndim * 2`, where ndim is a number of free
        parameters.

        :param nwalkers: int; The number of walkers in the ensemble. Minimum is 2 * number of free parameters.
        :param ndim: int; number of free variables
        :raise: RuntimeError; when condition `nwalkers < ndim * 2` is not satisfied
        """
        if nwalkers < ndim * 2:
            msg = f'Fit cannot be executed with fewer walkers ({nwalkers}) than twice the number of dimensions ({ndim})'
            raise RuntimeError(msg)



import numpy as np
from copy import copy

from . import (
    utils as crv_utils,
    c_managed,
    c_appx_router
)
from .. import (
    dynamic,
    surface
)
from .. container import OrbitalPositionContainer
from ... import const
from ... observer.mp_manager import manage_observations
from ... import settings
from ... logger import getLogger


logger = getLogger('binary_system.curves.c_router')


def resolve_curve_method(system, curve: str):
    """
    Resolves which curve calculating method to use based on the properties of the BinarySystem.

    :param curve: str; choose one of `lc` or `rv`
    :param system: elisa.binary_system.BinarySystem;
    :return: callable, curve calculating method
    """
    if curve == 'lc':
        fn_array = (
            getattr(system, '_compute_circular_synchronous_lightcurve'),
            getattr(system, '_compute_circular_spotty_asynchronous_lightcurve'),
            getattr(system, '_compute_circular_pulsating_lightcurve'),
            getattr(system, '_compute_eccentric_spotty_lightcurve'),
            getattr(system, '_compute_eccentric_lightcurve')
        )
    elif curve == 'rv':
        fn_array = (
            getattr(system, '_compute_circular_synchronous_rv_curve'),
            getattr(system, '_compute_circular_spotty_asynchronous_rv_curve'),
            getattr(system, '_compute_circular_pulsating_rv_curve'),
            getattr(system, '_compute_eccentric_spotty_rv_curve'),
            getattr(system, '_compute_eccentric_rv_curve_no_spots')
        )
    else:
        raise ValueError('Invalid value of argument `curve`. Only `lc` and `rv` are allowed')

    is_circular = system.eccentricity == 0
    is_eccentric = 1 > system.eccentricity > 0
    asynchronous_spotty_p = system.primary.synchronicity != 1 and system.primary.has_spots()
    asynchronous_spotty_s = system.secondary.synchronicity != 1 and system.secondary.has_spots()
    asynchronous_spotty_test = asynchronous_spotty_p or asynchronous_spotty_s

    spotty_test_eccentric = system.primary.has_spots() or system.secondary.has_spots()

    if is_circular:
        if asynchronous_spotty_test:
            logger.debug('calculating curve for circular binary system with with asynchronous spotty components')
            return fn_array[1]
        elif system.has_pulsations():
            logger.debug('calculating curve for circular binary system with pulsations without asynchronous spots')
            return fn_array[2]
        else:
            logger.debug('Calculating curve for circular binary system without '
                         'pulsations and without asynchronous spotty components.')
            return fn_array[0]
    elif is_eccentric:
        if spotty_test_eccentric:
            logger.debug('calculating curve for eccentric binary system with spotty components')
            return fn_array[3]
        else:
            logger.debug('calculating curve for eccentric binary system without spotty components')
            return fn_array[4]

    raise NotImplementedError("Orbit type not implemented or invalid.")


def prep_initial_system(binary, **kwargs):
    """
    Prepares base binary system from which curves will be calculated in case of circular synchronous binaries.

    :param binary: elisa.binary_system.system.BinarySystem
    :return: elisa.binary_system.container.OrbitalPositionContainer
    """
    from_this = dict(binary_system=binary, position=const.Position(0, 1.0, 0.0, 0.0, 0.0))
    initial_system = OrbitalPositionContainer.from_binary_system(**from_this)
    do_pulsations = kwargs.get('build_pulsations', True)
    initial_system.build(components_distance=1.0, build_pulsations=do_pulsations)
    return initial_system


def produce_circular_sync_curves(binary, initial_system, phases, curve_fn, crv_labels, **kwargs):
    """
    Auxiliary function to produce curve from circular synchronous binary system.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param initial_system: elisa.binary_system.container.OrbitalPositionContainer;
    :param phases: numpy.array;
    :param curve_fn: callable; function to calculate given type of the curve
    :param crv_labels: List; labels of the calculated curves (passbands, components,...)
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float
        * ** position_method** * - function definition; to evaluate orbital positions
        * ** phases ** * - numpy.array

    :return: Dict; calculated curves
    """

    crv_utils.prep_surface_params(initial_system, return_values=False, write_to_containers=True, **kwargs)
    fn = c_managed.produce_circ_sync_curves_mp
    fn_args = (binary, initial_system, crv_labels, curve_fn)
    return manage_observations(fn=fn, fn_args=fn_args, position=phases, **kwargs)


def produce_circular_spotty_async_curves(binary, curve_fn, crv_labels, **kwargs):
    """
    Function returns curve of assynchronous systems with circular orbits and spots.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param curve_fn: callable; curve function
    :param crv_labels: labels of the calculated curves (passbands, components,...)
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa .observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float

    :return: Dict; curves
    """
    phases = kwargs.pop("phases")
    position_method = kwargs.pop("position_method")
    orbital_motion = position_method(input_argument=phases, return_nparray=False, calculate_from='phase')
    ecl_boundaries = dynamic.get_eclipse_boundaries(binary, 1.0)

    from_this = dict(binary_system=binary, position=const.Position(0, 1.0, 0.0, 0.0, 0.0))
    initial_system = OrbitalPositionContainer.from_binary_system(**from_this)

    points = dict()
    for component in settings.BINARY_COUNTERPARTS:
        star = getattr(initial_system, component)
        _a, _c, _d = surface.mesh.mesh_detached(initial_system, 1.0, component, symmetry_output=True)
        points[component] = _a
        setattr(star, "points", copy(_a))
        setattr(star, "base_symmetry_points_number", _c)
        setattr(star, "inverse_point_symmetry_matrix", _d)

    fn_args = binary, initial_system, points, ecl_boundaries, crv_labels, curve_fn
    fn = c_managed.produce_circ_spotty_async_curves_mp
    return manage_observations(fn=fn, fn_args=fn_args, position=orbital_motion, **kwargs)


def produce_circular_pulsating_curves(binary, initial_system, phases, curve_fn, crv_labels, **kwargs):
    """
    Function returns curve of pulsating systems with circular orbits.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param initial_system: elisa.binary_system.container.OrbitalPositionContainer;
    :param phases: numpy.array;
    :param curve_fn: callable; curve function
    :param crv_labels: labels of the calculated curves (passbands, components,...)
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa .observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float

    :return: Dict; curves
    """
    fn = c_managed.produce_circ_pulsating_curves_mp
    fn_args = (binary, initial_system, crv_labels, curve_fn)
    return manage_observations(fn=fn, fn_args=fn_args, position=phases, **kwargs)


def produce_ecc_curves_no_spots(binary, curve_fn, crv_labels, **kwargs):
    """
    Function for generating curves of binaries with eccentric orbit and no spots where different curve
    integration approximations are evaluated and performed.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param curve_fn: generator function of curve point
    :param crv_labels: labels of the calculated curves (passbands, components,...)
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa .observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float

    :return: Dict; curves
    """
    phases = kwargs.pop("phases")

    # this condition checks if even to attempt to utilize apsidal line symmetry approximations
    # curve has to have enough point on orbit and have to span at least in 0.8 phase

    # this will remove large gap in phases
    max_diff = np.max(np.diff(np.sort(phases), n=1))
    phases_span_test = np.max(phases) - np.min(phases) - max_diff >= 0.79

    position_method = kwargs.pop("position_method")
    try_to_find_appx = c_appx_router.look_for_approximation(not binary.has_pulsations())

    _args = binary, phases, position_method, try_to_find_appx, phases_span_test, crv_labels, curve_fn
    appx_uid, run = c_appx_router.resolve_ecc_approximation_method(*_args, **kwargs)

    logger_messages = {
        'zero': 'curve will be calculated in a rigorous `phase to phase manner` without approximations',
        'one': 'one half of the curve points on the one side of the apsidal line will be interpolated',
        'two': 'geometry of the stellar surface on one half of the apsidal '
               'line will be copied from their closest symmetrical counterparts',
        'three': 'surface geometry at some orbital positions will not be recalculated from scratch due to '
                 'similarities to previous orbital positions instead the previous most similar shape will be used'
    }
    logger.info(logger_messages.get(appx_uid))
    return run()


def produce_ecc_curves_with_spots(binary, curve_fn, crv_labels, **kwargs):
    """
    Function for generating curves of binaries with eccentric orbit and spots.

    :param binary: elisa.binary_system.system.BinarySystem;
    :param curve_fn: curve generator function
    :param crv_labels: labels of the calculated curves (passbands, components,...)
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa .observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float

    :return: Dict; curves
    """
    phases = kwargs.pop("phases")
    position_method = kwargs.pop("position_method")
    orbital_motion = position_method(input_argument=phases, return_nparray=False, calculate_from='phase')

    potentials = binary.correct_potentials(phases, component="all", iterations=2)

    # pre-calculate the longitudes of each spot for each phase
    spots_longitudes = dynamic.calculate_spot_longitudes(binary, phases, component="all")
    fn_args = (binary, potentials, spots_longitudes, crv_labels, curve_fn)
    fn = c_managed.integrate_eccentric_curve_exactly
    return manage_observations(fn=fn, fn_args=fn_args, position=orbital_motion, **kwargs)



import numpy as np

from ... import (
    utils,
    const,
    umpy as up
)
from ... base.types import INT, FLOAT
from ... logger import getLogger
from ... binary_system.orbit.transform import OrbitProperties
from ... base.orbit.orbit import AbstractOrbit

logger = getLogger('binary_system.orbit.orbit')


def angular_velocity(period, eccentricity, distance):
    """
    Compute angular velocity for given components distance.
    This can be derived from facts that::


        w = dp/dt

        P * 1/2 * dp/dt = pi * a * b

        e = sqrt(1 - (b/a)^2)

    where a, b are respectively semi major and semi minor axis, P is period and e is eccentricity.


    :param period: float;
    :param eccentricity: float;
    :param distance: float;
    :return: float;
    """
    return ((2.0 * up.pi) / (period * 86400.0 * (distance ** 2))) * up.sqrt(
        (1.0 - eccentricity) * (1.0 + eccentricity))  # $\rad.sec^{-1}$


def primary_orbital_speed(m1, m2, a_red, components_distance):
    """
    Returns orbital speed of primary component with respect to the system centre of mass.

    :param m1: float; primary mass
    :param m2: float; secondary mass
    :param a_red: float; semi major axis of the primary component with respect to the system centre of mass
    :param components_distance: float;
    :return: float;
    """
    m = m1 + m2
    return m2 * np.sqrt((const.G / m) * ((2 / components_distance) - (m2 / (a_red * m))))


def velocity_vector_angle(eccentricity, true_anomaly):
    """
    Returns sine and cosine of angle between velocity vector and join vector.

    :param eccentricity: float;
    :param true_anomaly: float;
    :return: Tuple;
    """
    den = np.sqrt(1 + eccentricity**2 + 2 * eccentricity * np.cos(true_anomaly))
    sin = (1 + eccentricity * np.cos(true_anomaly)) / den
    cos = - (eccentricity * np.sin(true_anomaly)) / den
    return sin, cos


def create_orb_vel_vectors(system, components_distance):
    """
    Returns orbital velocity vectors for both components in reference frame of centre of mass.

    :param system: elisa.binary_system.container;
    :param components_distance: float;
    :return: float;
    """
    a_red = system.semi_major_axis * system.mass_ratio / (1 + system.mass_ratio)

    speed = primary_orbital_speed(system.primary.mass, system.secondary.mass, a_red,
                                  system.semi_major_axis * components_distance)

    sin, cos = velocity_vector_angle(system.eccentricity, system.position.true_anomaly)

    velocity = {'primary': np.array([cos * speed, -sin * speed, 0])}
    velocity['secondary'] = - velocity['primary'] / system.mass_ratio

    return velocity


def distance_to_center_of_mass(primary_mass, secondary_mass, distance):
    """
    Return distance from primary and from secondary component to center of mass.

    :param primary_mass: float
    :param secondary_mass: float
    :param distance: Union[float, numpy.array]
    :return: Tuple[Union[float, numpy.array]];
    """
    mass = primary_mass + secondary_mass
    com_from_primary = (distance * secondary_mass) / mass
    return com_from_primary, distance - com_from_primary


def orbital_semi_major_axes(r, eccentricity, true_anomaly):
    """
    Return orbital semi major axis from component distance, eccentricity and true anomaly.

    :param r: float or numpy.array; distance from center of mass to object
    :param eccentricity: float or numpy.array; orbital eccentricity
    :param true_anomaly: float or numpy.array; true anomaly of orbital motion
    :return: Union[float, numpy.array]
    """
    return r * (1.0 + eccentricity * up.cos(true_anomaly)) / (1.0 - up.power(eccentricity, 2))


def component_distance_from_mean_anomaly(eccentricity, true_anomaly):
    """
    Return component distance in SMA from semi-major axis, eccentricity and true anomaly.

    :param r: float or numpy.array; distance from center of mass to object
    :param eccentricity: float or numpy.array; orbital eccentricity
    :param true_anomaly: float or numpy.array; true anomaly of orbital motion
    :return: Union[float, numpy.array]
    """
    return (1.0 - up.power(eccentricity, 2)) / (1.0 + eccentricity * up.cos(true_anomaly))


def get_approx_ecl_angular_width(forward_radius1, forward_radius2, components_distance, inclination):
    """
    Returns angular width of the eclipse assuming spherical components.

    :param forward_radius1: float;
    :param forward_radius2: float;
    :param components_distance: float; in SMA
    :param inclination: float; in radians
    :return: tuple; angular half-width of the eclipse and the inner plateau
    """
    # tilt of the orbital plane and z-axis in the observer reference frame
    tilt = np.abs(const.HALF_PI - inclination)
    # maximum apparent distance between components where eclipse is possible
    r_outer = forward_radius1 + forward_radius2
    r_inner = np.abs(forward_radius1 - forward_radius2)
    # closest aparent distances of component centres
    r_close = components_distance * np.sin(tilt)

    # checking if eclipses occur

    nu_outer = 0.0 if r_close >= r_outer else \
        np.arcsin(np.sqrt(
            np.power(r_outer/components_distance, 2) - np.power(np.sin(tilt), 2)
        ))
    nu_inner = 0.0 if r_close >= r_inner else \
        np.arcsin(np.sqrt(
            np.power(r_inner / components_distance, 2) - np.power(np.sin(tilt), 2)
        ))

    return nu_outer, nu_inner


class Orbit(AbstractOrbit):
    """
    Object representing orbit of a binary system. Accessible as an attribute of an BinarySystem object

    Input parameters:

    :param period: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]; Orbital period of binary
                   star system. If unit is not specified, default period unit is assumed (days).
    :param inclination: ; Union[float, astropy.units.Quantity]; If unitless values is supplied, default unit
                          suppose to be radians.
    :param eccentricity: Union[(numpy.)int, (numpy.)float];
    :param argument_of_periastron: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]; If unit is
                                   not supplied, value in radians is assumed.
    :param phase_shift: float;

    Output parameters:

    :periastron_distance: float; Distance in periastron (in unit of semi major axis (a = a1 + a2) of relative motion)
    :periastron_phase: float; true (not photometric but computed from mean anomaly) phase of periastron.
    """

    MANDATORY_KWARGS = ['period', 'inclination', 'eccentricity', 'argument_of_periastron']
    OPTIONAL_KWARGS = ['phase_shift']
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS

    def __init__(self, **kwargs):
        utils.invalid_kwarg_checker(kwargs, Orbit.ALL_KWARGS, Orbit)
        utils.check_missing_kwargs(self.__class__.MANDATORY_KWARGS, kwargs, instance_of=self.__class__)
        kwargs = OrbitProperties.transform_input(**kwargs)

        super(Orbit, self).__init__(**kwargs)

        # default valeus of properties
        self.period = np.nan
        self.eccentricity = np.nan
        self.argument_of_periastron = np.nan
        self.inclination = np.nan

        self.periastron_distance = np.nan
        self.periastron_phase = np.nan
        self.semi_major_axis = np.nan
        self.phase_shift = 0.0

        # values of properties
        logger.debug(f"setting properties of orbit")
        for kwarg in kwargs:
            setattr(self, kwarg, kwargs[kwarg])

        self.periastron_distance = self.compute_periastron_distance()
        self.conjunctions = self.get_conjuction()
        self.periastron_phase = - self.conjunctions["primary_eclipse"]["true_phase"] % 1

    @classmethod
    def phase_to_mean_anomaly(cls, phase):
        """
        returns mean anomaly of points on orbit as a function of phase

        :param phase: Union[numpy.array, float];
        :return: Union[numpy.array, float];
        """
        return const.FULL_ARC * phase

    @classmethod
    def mean_anomaly_to_phase(cls, mean_anomaly):
        """
        returns phase of points on orbit as a function of mean anomaly

        :param mean_anomaly: Union[numpy.array, float];
        :return:  Union[numpy.array, float];
        """
        return mean_anomaly / const.FULL_ARC

    def mean_anomaly_fn(self, eccentric_anomaly: float, *args) -> float:
        """
        Definition of Kepler equation for scipy _solver in Orbit.eccentric_anomaly.

        :param eccentric_anomaly: float;
        :param args: Tuple; (mean_anomaly, )
        :return: float;
        """
        mean_anomaly, = args
        return eccentric_anomaly - self.eccentricity * up.sin(eccentric_anomaly) - mean_anomaly

    def mean_anomaly_to_eccentric_anomaly(self, mean_anomaly: float) -> float:
        """
        Solves Kepler equation for eccentric anomaly via mean anomaly.

        :param mean_anomaly: float;
        :return: float;
        """
        import scipy.optimize
        try:
            solution = scipy.optimize.newton(self.mean_anomaly_fn, 1.0, args=(mean_anomaly, ), tol=1e-10)
            if not up.isnan(solution):
                if solution < 0:
                    solution += const.FULL_ARC
                return solution
            else:
                return False
        except Exception as e:
            logger.debug(f"solver scipy.optimize.newton in function Orbit.eccentric_anomaly did not provide "
                         f"solution.\n Reason: {e}")
            return False

    def eccentric_anomaly_to_mean_anomaly(self, eccentric_anomaly):
        """
        Returns mean anomaly as a function of eccentric anomaly calculated using Kepler equation.

        :param eccentric_anomaly: numpy.array;
        :return: numpy.array
        """
        return (eccentric_anomaly - self.eccentricity * up.sin(eccentric_anomaly)) % const.FULL_ARC

    def eccentric_anomaly_to_true_anomaly(self, eccentric_anomaly):
        """
        Returns true anomaly as a function of eccentric anomaly and eccentricity.

        :param eccentric_anomaly: Union[numpy.array, float]
        :return: Union[numpy.array, float]
        """
        true_anomaly = 2.0 * up.arctan(
            up.sqrt((1.0 + self.eccentricity) / (1.0 - self.eccentricity)) * up.tan(eccentric_anomaly / 2.0))
        true_anomaly[true_anomaly < 0] += const.FULL_ARC
        return true_anomaly

    def true_anomaly_to_eccentric_anomaly(self, true_anomaly):
        """
        Returns eccentric anomaly as a function of true anomaly and eccentricity.

        :param true_anomaly: Union[numpy.array, float];
        :return: Union[numpy.array, float];
        """
        eccentric_anomaly = \
            2.0 * up.arctan(up.sqrt((1.0 - self.eccentricity) / (1.0 + self.eccentricity)) * up.tan(true_anomaly / 2.0))
        eccentric_anomaly[eccentric_anomaly < 0] += const.FULL_ARC
        return eccentric_anomaly

    def relative_radius(self, true_anomaly):
        """
        Calculates the length of radius vector of elipse where a=1.

        :param true_anomaly: Union[numpy.array, float];
        :return: Union[numpy.array, float];
        """
        return (1.0 - self.eccentricity ** 2) / (1.0 + self.eccentricity * up.cos(true_anomaly))

    def true_anomaly_to_azimuth(self, true_anomaly):
        """
        Convert true anomaly angle to azimuth angle measured from y axis in 2D plane.

        ::

            azimuth 0 alligns with y axis
                     | 0
                pi/2 |
                -----------
                     |
                     |

        :param true_anomaly: Union[numpy.array, float];
        :return: Union[numpy.array, float];
        """
        return (true_anomaly + self.argument_of_periastron) % const.FULL_ARC

    def azimuth_to_true_anomaly(self, azimuth):
        """
        Calculates the azimuth form given true anomaly

        :param azimuth: Union[numpy.array, float]
        :return: Union[numpy.array, float]
        """
        return (azimuth - self.argument_of_periastron) % const.FULL_ARC

    def orbital_motion(self, phase):
        """
        Function takes photometric phase of the binary system as input and calculates positions of the secondary
        component in the frame of reference of primary component.

        :param phase: Union[numpy.array, float]; photometric phases
        :return: numpy.array; matrix consisting of column stacked vectors distance, azimut angle, true anomaly and phase

        ::

            numpy.array((r1, az1, nu1, phs1),
                        (r2, az2, nu2, phs2),
                         ...       
                        (rN, azN, nuN, phsN))
        """
        # ability to accept scalar as input
        if isinstance(phase, (int, INT, float, FLOAT)):
            phase = np.array([FLOAT(phase)])
        # photometric phase to phase measured from periastron
        true_phase = self.true_phase(phase=phase, phase_shift=self.conjunctions['primary_eclipse']['true_phase'])

        mean_anomaly = self.phase_to_mean_anomaly(phase=true_phase)
        eccentric_anomaly = np.array([self.mean_anomaly_to_eccentric_anomaly(mean_anomaly=xx)
                                      for xx in mean_anomaly])
        true_anomaly = self.eccentric_anomaly_to_true_anomaly(eccentric_anomaly=eccentric_anomaly)
        distance = self.relative_radius(true_anomaly=true_anomaly)
        azimut_angle = self.true_anomaly_to_azimuth(true_anomaly=true_anomaly)

        return np.column_stack((distance, azimut_angle, true_anomaly, phase))

    def orbital_motion_from_azimuths(self, azimuth):
        """
        Function takes azimuths of the binary system (angle between ascending node (y) as input and calculates
        positions of the secondary component in the frame of reference of primary component.

        :param azimuth: Union[numpy.array, float];
        :return: numpy.array; matrix consisting of column stacked vectors distance,
                              azimut angle, true anomaly and phase

        ::

               numpy.array((r1, az1, ni1, phs1),
                           (r2, az2, ni2, phs2),
                            ...
                           (rN, azN, niN, phsN))

        """
        true_anomaly = self.azimuth_to_true_anomaly(azimuth)
        distance = self.relative_radius(true_anomaly=true_anomaly)
        eccentric_anomaly = self.true_anomaly_to_eccentric_anomaly(true_anomaly)
        mean_anomaly = self.eccentric_anomaly_to_mean_anomaly(eccentric_anomaly)
        true_phase = self.mean_anomaly_to_phase(mean_anomaly)
        phase = self.phase(true_phase, phase_shift=self.conjunctions['primary_eclipse']['true_phase'])
        return np.column_stack((distance, azimuth, true_anomaly, phase))

    def get_conjuction(self):
        """
        Compute and return photometric phase of conjunction (eclipses).
        We assume that primary component is placed in center of coo system and observation unit vector is [1, 0, 0]

        return dictionary is in shape::

            {
                type_of_eclipse <`primary_eclipse` or `secondary_eclipse`>: {
                    'true_phase': float,
                    'true_anomaly': float,
                    'mean_anomaly': float,
                    'eccentric_anomaly'float:
                },
                ...
            }

        :return: Dict;
        """
        # determining order of eclipses
        conjuction_arc_list = []
        try:
            if 0 <= self.inclination <= const.PI / 2.0:
                conjuction_arc_list = [const.PI / 2.0, 3.0 * const.PI / 2.0]
            elif const.PI / 2.0 < self.inclination <= const.PI:
                conjuction_arc_list = [3.0 * const.PI / 2.0, const.PI / 2.0]
        except Exception as e:
            raise TypeError(f'Invalid type of {self.__class__.__name__}.inclination - {str(e)}.')

        conjunction_quantities = dict()
        for alpha, idx in list(zip(conjuction_arc_list, ['primary_eclipse', 'secondary_eclipse'])):
            # true anomaly of conjunction (measured from periastron counter-clokwise)
            true_anomaly_of_conjuction = self.azimuth_to_true_anomaly(alpha)  # \nu_{con}

            # eccentric anomaly of conjunction (measured from apse line)
            eccentric_anomaly_of_conjunction = (2.0 * up.arctan(
                up.sqrt((1.0 - self.eccentricity) / (1.0 + self.eccentricity)) *
                up.tan(true_anomaly_of_conjuction / 2.0))) % const.FULL_ARC

            # mean anomaly of conjunction (measured from apse line)
            mean_anomaly_of_conjunction = (eccentric_anomaly_of_conjunction -
                                           self.eccentricity *
                                           up.sin(eccentric_anomaly_of_conjunction)) % const.FULL_ARC

            # true phase of conjunction (measured from apse line)
            true_phase_of_conjunction = (mean_anomaly_of_conjunction / const.FULL_ARC) % 1.0

            conjunction_quantities[idx] = dict()
            conjunction_quantities[idx]["true_anomaly"] = true_anomaly_of_conjuction
            conjunction_quantities[idx]["eccentric_anomaly"] = eccentric_anomaly_of_conjunction
            conjunction_quantities[idx]["mean_anomaly"] = mean_anomaly_of_conjunction
            conjunction_quantities[idx]["true_phase"] = true_phase_of_conjunction

        return conjunction_quantities

    def compute_periastron_distance(self):
        """
        Calculates relative periastron distance in SMA units.

        :return: float;
        """
        periastron_distance = self.relative_radius(true_anomaly=np.array([0])[0])
        logger.debug(f"setting property periastron_distance "
                     f"of class instance {self.__class__.__name__} to {periastron_distance}")
        return periastron_distance
